/* master include file for internal database definitions */

/* if CLEANUP is defined, then we will free all of our data structures,
   otherwise we just rely on the operating system to do it. The only time
   we would need to clean up properly is if the entire database package
   were being called as a subroutine from another program which would want
   to be able to reuse the space after the database stuff is all done. */

#undef  CLEANUP

/* if PARANOID is defined, we will check all sorts of things. These checks
   are probably only useful for people developing new applications. This
   flag is off for CP/M so the fool thing will still fit. */

#undef  PARANOID

#define VERSION	21	/* this version number */
#define VARIANT "a"	/* variant of version */
#define	BLKSIZE	1024	/* size of blocks used for data and indices */
#define SYSSIZE 128	/* size of block used by low-level I/O guys */
/* NOTE: some code assumes that BLKSIZE is < 64K */
#define INOSIZE	128	/* size of sub-blocks used for inodes */
/* NOTE: some code assumes BLKSIZE / INOSIZE is a power of 2 */
#define MAXFNAM 30	/* maximum length of field name (not including 0) */
#define MAXKEY	330	/* maximum length of a key field */
/* MAXKEY = (BLKSIZE - ((sizeof(bool) + sizeof(word) + sizeof(TUPPTR))) / 3 -
    sizeof(TUPPTR) - sizeof(POINTER) */
/* MUST be even because of the 68K round up in INDEX1.C */

#ifdef XENIX
#define COPY(dest, source, length)	movmem(source, dest, length)
#endif
#ifdef MSDOS
#define COPY(dest, source, length)	movmem(source, dest, length)
#endif
#ifdef CPM
#define COPY(dest, source, length)	blockmv(dest, source, length)
#endif

/* use this typedef so we can distinguish when we are using a char as
   a numeric quantity: */

#ifdef XENIX
typedef unsigned char byte;
#endif
#ifdef MSDOS
typedef unsigned char byte;
#endif
#ifdef CPM
typedef char byte;		/* dumb compiler doesn't understand unsigned
				   char, but chars are unsigned anyway */
#endif

/* use this typedef and 'true' and 'false' to fudge proper booleans: */

typedef char bool;
#define true	1
#define false	0

/* use this typedef so we can change to 32 bit if needed: */

typedef unsigned short word;

/* define a 'nil' suitable for use with pointers. This definition should
   really be dependent on the sizeof pointers, but this compiler can't do
   a #if to handle that. This value should be changed to '0l' for versions
   in which pointers are the same size as longs instead of ints. */

#ifdef M68K
#define nil	0
#endif
#ifdef M8086
#define nil	0l
#endif
#ifdef M8080
#define nil	0
#endif

/* internal structure used for fields in the dictionary: */

typedef struct {
    char *f_name;
    word f_len;			/* used only for FTCHAR, FTSTRING fields */
    byte f_type;		/* type of field - see below */
} FIELD;

/* definitions of field types: */

#define FTBOOL	0	/* a boolean field */
#define FTINT	1	/* an integer (16 or 32 bit) field */
#define FTLONG	2	/* a long integer (32 bit) field */
#define FTFLOAT	3	/* a floating point (double in C) field */
#define FTCHAR	4	/* a field of some given length */
#define FTSTRING 5	/* a character string of some maximum length */
#define FTUCSTR	6	/* characters string for which case is ignored */
#define FTSHORT 7	/* a short integer (16 bit) field */
#define FTLAST	7	/* id of last valid field type */

/* internal structure representing a dictionary: */

typedef struct {
    FIELD *d_flds;		/* pointer to array of FIELD's */
    int d_fldCnt;		/* number of FIELD's in dictionary */
} DICT;

/* structure of a pointer used within indices and i-nodes: */
/*   (a null pointer has the file number = 0) */

typedef struct {
#ifdef M68K
    byte p_pad;			/* pad byte (0xff) */
    byte p_file;		/* file number */
    word p_block;		/* 128 byte block number */
#else
    word p_block;		/* 128 byte block number */
    byte p_file;		/* file number */
    byte p_pad;			/* pad byte (0xff) */
#endif
} POINTER;

/* internal structure of an index: */

typedef struct index {
    struct index *x_next;	/* next in chain for this relation */
    int x_field;		/* single field which is the key */
    word x_len;			/* length of key */
    POINTER x_root;		/* the root record */
    byte x_type;		/* type of key */
    bool x_unique;		/* true if no duplicates allowed */
} INDEX;

/* internal stucture of a database: */

typedef struct database {
    char *db_name;		/* name of this database */
    struct database *db_next;	/* next in chain of active databases */
    DICT db_dict;		/* dictionary for this database */
    bool db_act;		/* false if database is closed */
} DATABASE;

/* structure of a field descriptor with a relation. */

typedef struct {
    long rf_pos;		/* pos of field in tuple */
    int rf_field;		/* dictionary field number (0-origin) */
    word rf_scnCnt;		/* number of active scans on this field */
    bool rf_index;		/* true if field is in an index */
    bool rf_mod;		/* true if field modified in this tuple */
} RFIELD;

/* internal structure of a relation: */

typedef struct relation {
    struct relation *r_next;	/* next in chain of open relations */
    INDEX *r_index;		/* chain of indexes on this relation */
    RFIELD *r_fields;		/* pointer to array of field numbers */
    DATABASE *r_dbptr;		/* pointer to parent database */
    byte *r_riol;		/* pointer to RIOL structure in DBIO.C */
    char *r_name;		/* name of this relation */
    long r_len;			/* total length of a tuple of this relation*/
    long r_tnext;		/* next free tuple */
    long r_tfree;		/* free list of tuples */
    long r_tcurr;		/* current tuple */
    int r_fldCnt;		/* number of fields in this relation */
    word r_dbase;		/* number of parent database */
    word r_fSize;		/* max file size in K for this relation */
    POINTER r_xnext;		/* next unused block of index/inode */
    POINTER r_xfree;		/* free list of index/inode blocks */
    POINTER r_nfree;		/* free list of smaller inode blocks */
    bool r_act;			/* false if relation is closed */
    bool r_change;		/* true if relation header been changed */
} RELATION;
