#include "system.h"
#include "dbint.h"

extern d_error();

extern DATABASE *d_dpget2();
extern RELATION *d_rpget2();
extern d_pfLog();

extern byte d_bdGet();
extern d_bgStuff();
extern d_bgStr();
extern d_bdPut();
extern d_bpStuff();

extern d_putTup();
extern d_uIndF();
extern d_xsMod();

RELATION *d_FRelPtr;

/* routines for dealing with individual fields */

/*
 * fLookup - get a field number corresponding to the passed field name.
 *	     The field number is w.r.t. the dictionary, so to get a field
 *	     number w.r.t. a relation, this value must be looked up in that
 *	     relation. Return 0 if not found, else 1-origin field number.
 */

int
fLookup(dbNum, name)
    word dbNum;
    char *name;
{
    DATABASE *db;
    FIELD *f;
    int n;

    db = d_dpget2(dbNum);
    f = db->db_dict.d_flds;
    n = 0;
    while (n != db->db_dict.d_fldCnt && strcmp(name, f++->f_name) != 0) {
	n++;
    }
    return((n == db->db_dict.d_fldCnt) ? 0 : (n + 1));
}

/*
 * dFldCnt - return the number of fields in the given database dictionary.
 */

int
dFldCnt(dbNum)
    word dbNum;
{

    return(d_dpget2(dbNum)->db_dict.d_fldCnt);
}

/*
 * fData - get field data corresponding to the passed field number.
 *	   The field number is w.r.t. the dictionary. A pointer to the
 *	   field name (not a copy) is returned.
 */

char *
fData(dbNum, fldNum, pType, pLen)
    word dbNum;
    int fldNum;
    byte *pType;
    word *pLen;
{
    DATABASE *db;
    FIELD *f;

    db = d_dpget2(dbNum);
    if (fldNum <= 0 || fldNum > db->db_dict.d_fldCnt) {
	return(nil);
    } else {
	f = &db->db_dict.d_flds[fldNum - 1];
	*pLen = f->f_len;
	*pType = f->f_type;
	return(f->f_name);
    }
}

/*
 * rFldCnt - return the number of fields in the given relation.
 */

int
rFldCnt(relNum)
    word relNum;
{

    return(d_rpget2(relNum)->r_fldCnt);
}

/*
 * fRtoD - take a relation id and a field id within the relation, and return
 *	   the corresponding dictionary field number.
 */

int
fRtoD(relNum, fldNum)
    word relNum;
    int fldNum;
{

    d_FRelPtr = d_rpget2(relNum);
    if (fldNum <= 0 || fldNum > d_FRelPtr->r_fldCnt) {
	return(0);
    } else {
	return(d_FRelPtr->r_fields[fldNum - 1].rf_field + 1);
    }
}

/* 
 * fDtoR - take a dictionary field number and a relation, and find the
 *	   the relation-relative field number. Return 0 if none found,
 *	   else return the 1-origin field number.
 */

int
fDtoR(relNum, fldNum)
    word relNum;
    int fldNum;
{
    RFIELD *rf;
    int cnt;

    if (fldNum-- <= 0) {
	d_error(23);
    }
    d_FRelPtr = d_rpget2(relNum);
    cnt = d_FRelPtr->r_fldCnt;
    rf = d_FRelPtr->r_fields;
    while (cnt != 0 && rf++->rf_field != fldNum) {
	cnt--;
    }
    if (cnt == 0) {
	d_error(23);
    }
    return(d_FRelPtr->r_fldCnt - cnt + 1);
}

/*
 * relDB - return the database identifier for a relation.
 */

word
relDB(relNum)
    word relNum;
{

    return(d_rpget2(relNum)->r_dbase);
}

/*
 * d_chFld - check a field for validity in a relation. At the same time,
 *	     check the passed tuple number, and type. Also, based on 'mod',
 *	     either get or put the field. If the field is put, check for and
 *	     do any required logging.
 */

d_chFld(relNum, tuple, fNum, fType, mod, ptr)
    word relNum;
    long tuple;
    int fNum;
    byte fType;
    bool mod;
    byte *ptr;
{
    FIELD *f;
    RFIELD *rf;
    long pos;
    word len, sLen;

    d_FRelPtr = d_rpget2(relNum);
    if (tuple < 1l || tuple >= d_FRelPtr->r_tnext) {
	d_error(24);
    }
    if (fNum <= 0 || fNum > d_FRelPtr->r_fldCnt) {
	d_error(23);
    }
    pos = (tuple - 1l) * d_FRelPtr->r_len;
    if (d_bdGet(d_FRelPtr, pos)) {
	d_error(16);
    }
    rf = &d_FRelPtr->r_fields[fNum - 1];
    f = &d_FRelPtr->r_dbptr->db_dict.d_flds[rf->rf_field];
    if (f->f_type == FTUCSTR ? fType != FTSTRING : fType != f->f_type) {
	d_error(25);
    }
    if (tuple != d_FRelPtr->r_tcurr) {
	d_putTup(d_FRelPtr);
	d_xsMod(d_FRelPtr, false);
	d_FRelPtr->r_tcurr = tuple;
    }
    if (mod && rf->rf_index && !rf->rf_mod) {
	d_uIndF(d_FRelPtr, fNum - 1);
	rf->rf_mod = true;
    }
    pos += rf->rf_pos;
    len = f->f_len;
    if (mod) {
	if (fType == FTSTRING) {
	    /* we do it this way for strings, since if we are given a short
	       actual string for a long string field, things go faster if
	       we only put as many bytes as are in the actual string. We
	       also have to check to make sure that we don't write MORE
	       bytes than will fit, and that we always have a 0 at the end
	       of the stored string. */
	    sLen = strlen(ptr);
	    if (sLen >= len) {
		sLen = len - 1;
	    }
	    d_bpStuff(d_FRelPtr, pos, ptr, sLen);
	    d_bdPut(d_FRelPtr, pos + sLen, 0);
	} else {
	    d_bpStuff(d_FRelPtr, pos, ptr, len);
	}
	/* NOTE: we do the log as the very last thing in this routine, so
	   that we don't disallow recursion (i.e. using another database
	   as part of the logging procedure). */
	d_pfLog(relNum, tuple, fNum, fType, len, ptr);
    } else {
	d_bgStuff(d_FRelPtr, pos, ptr, len);
    }
}

/*
 * gfBool - get a boolean field from a record.
 */

gfBool(relNum, tuple, fNum, pRes)
    word relNum;
    long tuple;
    int fNum;
    bool *pRes;
{
    d_chFld(relNum, tuple, fNum, FTBOOL, false, pRes);
}

/*
 * gfShort - get a 16 bit field from a record.
 */

gfShort(relNum, tuple, fNum, pRes)
    word relNum;
    long tuple;
    int fNum;
    short *pRes;
{
    d_chFld(relNum, tuple, fNum, FTSHORT, false, pRes);
}

/*
 * gfInt - get an integer field from a record.
 *         Actual size of the object depends on the size of an int on
 *         this machine.
 */

gfInt(relNum, tuple, fNum, pRes)
    word relNum;
    long tuple;
    int fNum;
    int *pRes;
{
    if (sizeof(int) == sizeof(long)) {
	gfLong(relNum, tuple, fNum, pRes);
    } else {
	gfShort(relNum, tuple, fNum, pRes);
    }
}

/*
 * gfLong - get a 32 bit field from a record.
 */

gfLong(relNum, tuple, fNum, pRes)
    word relNum;
    long tuple;
    int fNum;
    long *pRes;
{
    d_chFld(relNum, tuple, fNum, FTLONG, false, pRes);
}

/*
 * gfFloat - get a double floating value from a record.
 */

gfFloat(relNum, tuple, fNum, pRes)
    word relNum;
    long tuple;
    int fNum;
    double *pRes;
{
    d_chFld(relNum, tuple, fNum, FTFLOAT, false, pRes);
}

/*
 * gfChar - get a character field from a record.
 */

gfChar(relNum, tuple, fNum, ptr)
    word relNum;
    long tuple;
    int fNum;
    byte *ptr;
{
    d_chFld(relNum, tuple, fNum, FTCHAR, false, ptr);
}

/*
 * gfString - get a string field from a tuple.
 */

gfString(relNum, tuple, fNum, ptr)
    word relNum;
    long tuple;
    int fNum;
    char *ptr;
{
    d_chFld(relNum, tuple, fNum, FTSTRING, false, ptr);
}

/*
 * pfBool - put a boolean value to a record.
 */

pfBool(relNum, tuple, fNum, val)
    word relNum;
    long tuple;
    int fNum;
    bool val;
{
    d_chFld(relNum, tuple, fNum, FTBOOL, true, &val);
}

/*
 * pfShort - put a short value to a record.
 */

pfShort(relNum, tuple, fNum, val)
    word relNum;
    long tuple;
    int fNum;
    short val;
{
    d_chFld(relNum, tuple, fNum, FTSHORT, true, &val);
}

/*
 * pfInt - put an integer field to a record.
 *         Actual size of the object depends on the size of an int on
 *         this machine.
 */

pfInt(relNum, tuple, fNum, val)
    word relNum;
    long tuple;
    int fNum;
    int val;
{
    if (sizeof(int) == sizeof(long)) {
	pfLong(relNum, tuple, fNum, val);
    } else {
	pfShort(relNum, tuple, fNum, val);
    }
}

/*
 * pfLong - put a long int value to a record.
 */

pfLong(relNum, tuple, fNum, val)
    word relNum;
    long tuple;
    int fNum;
    long val;
{
    d_chFld(relNum, tuple, fNum, FTLONG, true, &val);
}

/*
 * pfFloat - put a float value to a record.
 */

pfFloat(relNum, tuple, fNum, val)
    word relNum;
    long tuple;
    int fNum;
    double val;
{
    d_chFld(relNum, tuple, fNum, FTFLOAT, true, &val);
}

/*
 * pfChar - put a character value to a record.
 */

pfChar(relNum, tuple, fNum, ptr)
    word relNum;
    long tuple;
    int fNum;
    byte *ptr;
{
    d_chFld(relNum, tuple, fNum, FTCHAR, true, ptr);
}

/*
 * pfString - put a string value to a tuple.
 */

pfString(relNum, tuple, fNum, ptr)
    word relNum;
    long tuple;
    int fNum;
    char *ptr;
{
    d_chFld(relNum, tuple, fNum, FTSTRING, true, ptr);
}
