#include "system.h"
#include "dbint.h"

/* top level of index handling routines */

extern d_error();
extern d_setErr();
extern byte *d_alloc();

extern RELATION *d_rpget2();

extern d_xSetup();
extern d_xIns();
extern d_xDel();
extern long d_xFnd();

extern d_scDTup();

/*
 * d_xsMod - set all of the 'mod' flags in the fields of a relation.
 */

d_xsMod(r, val)
    RELATION *r;
    bool val;
{
    RFIELD *rf;
    word cnt;

    rf = &r->r_fields[0];
    for (cnt = r->r_fldCnt; cnt-- != 0;) {
	rf++->rf_mod = val;
    }
}

/*
 * d_putTup - routine called to re-index all changed fields of the current
 *	      tuple in the specified relation.
 */

d_putTup(r)
    RELATION *r;
{
    INDEX *x;
    RFIELD *rf;

    if (r->r_tcurr > 0l) {
	x = r->r_index;
	while (x != nil) {
	    if (x->x_field >= 0) {
		rf = &r->r_fields[x->x_field];
		if (rf->rf_mod) {
		    rf->rf_mod = false;
		    d_xSetup(r, x);
		    d_xIns(r->r_tcurr);
		}
	    }
	    x = x->x_next;
	}
	r->r_tcurr = 0l;
    }
}

/*
 * d_uIndF - unindex a specific field of the current tuple.
 */

d_uIndF(r, fNum)
    RELATION *r;
    int fNum;
{
    INDEX *x;

    x = r->r_index;
    while (x != nil) {
	if (x->x_field == fNum) {
	    d_scDTup(r->r_tcurr);
	    d_xSetup(r, x);
	    d_xDel(r->r_tcurr);
	}
	x = x->x_next;
    }
}

/*
 * d_uInd - unindex all fields of the current tuple.
 */

d_uInd(r)
    RELATION *r;
{
    INDEX *x;

    x = r->r_index;
    while (x != nil) {
	if (x->x_field >= 0) {
	    d_scDTup(r->r_tcurr);
	    d_xSetup(r, x);
	    d_xDel(r->r_tcurr);
	}
	x = x->x_next;
    }
}

/*
 * rUpdate - update the indices associated with the given relation.
 */

rUpdate(relNum)
    word relNum;
{
    RELATION *r;

    r = d_rpget2(relNum);
    d_putTup(r);
    d_xsMod(r, false);
}

/*
 * d_iFind - find an index for a given relation/field.
 */

INDEX *
d_iFind(relNum, fldNum, rp)
    word relNum;
    int fldNum;
    RELATION **rp;
{
    RELATION *r;
    INDEX *x;

    r = d_rpget2(relNum);
    if (fldNum == 0 || fldNum > r->r_fldCnt ||
	    fldNum < - r->r_dbptr->db_dict.d_fldCnt) {
	d_error(23);
    }
    if (fldNum > 0) {
	--fldNum;
    }
    x = r->r_index;
    while (x != nil && x->x_field != fldNum) {
	x = x->x_next;
    }
    *rp = r;
    return(x);
}

/*
 * isIndexed - check if a given field of a relation is indexed.
 */

bool
isIndexed(relNum, fldNum)
    word relNum;
    int fldNum;
{
    RELATION *r;

    return(d_iFind(relNum, fldNum, &r) != nil);
}

/*
 * isUnique - check if a given indexed field is indexed uniquely.
 */

bool
isUnique(relNum, fldNum)
    word relNum;
    int fldNum;
{
    RELATION *r;
    INDEX *x;

    x = d_iFind(relNum, fldNum, &r);
    return(x == nil ? false : x->x_unique);
}

/*
 * d_sChk0 - check a scan/find on a given relation/field.
 *	     This variant doesn't check the field type.
 */

INDEX *
d_sChk0(relNum, fldNum, rp)
    word relNum;
    int fldNum;
    RELATION **rp;
{
    INDEX *x;

    if (fldNum > 0) {
	rUpdate(relNum);
    }
    x = d_iFind(relNum, fldNum, rp);
    if (x == nil) {
	d_error(28);
    }
    return(x);
}

/*
 * d_sChk - check a scan/find on a given relation/field.
 *	    This variant checks field types as well.
 */

INDEX *
d_sChk(relNum, fldNum, rp, typ)
    word relNum;
    int fldNum;
    RELATION **rp;
    byte typ;
{
    INDEX *x;

    x = d_sChk0(relNum, fldNum, rp);
    if ((x->x_type == FTUCSTR) ? (typ != FTSTRING) : (x->x_type != typ)) {
	d_error(25);
    }
    return(x);
}

/*
 * d_foChk - common routine to check a 'fOne' and do it.
 */

long
d_foChk(relNum, fldNum, key, typ)
    word relNum;
    int fldNum;
    byte *key;
    byte typ;
{
    RELATION *r;
    INDEX *x;

    x = d_sChk(relNum, fldNum, &r, typ);
    d_xSetup(r, x);
    return(d_xFnd(key));
}

/*
 * user interface routines for finding some tuple with a given key.
 */

long
fOneBool(relNum, fldNum, valu)
    word relNum;
    int fldNum;
    bool valu;
{
    return(d_foChk(relNum, fldNum, &valu, FTBOOL));
}

long
fOneShort(relNum, fldNum, valu)
    word relNum;
    int fldNum;
    short valu;
{
    return(d_foChk(relNum, fldNum, &valu, FTSHORT));
}

long
fOneLong(relNum, fldNum, valu)
    word relNum;
    int fldNum;
    long valu;
{
    return(d_foChk(relNum, fldNum, &valu, FTLONG));
}

long
fOneInt(relNum, fldNum, valu)
    word relNum;
    int fldNum;
    int valu;
{
    return(d_foChk(relNum, fldNum, &valu,
		   sizeof(int) == sizeof(long) ? FTLONG : FTSHORT
		  )
    );
}

long
fOneFloat(relNum, fldNum, valu)
    word relNum;
    int fldNum;
    double valu;
{
    return(d_foChk(relNum, fldNum, &valu, FTFLOAT));
}

long
fOneChar(relNum, fldNum, valu)
    word relNum;
    int fldNum;
    byte *valu;
{
    return(d_foChk(relNum, fldNum, valu, FTCHAR));
}

long
fOneString(relNum, fldNum, valu)
    word relNum;
    int fldNum;
    char *valu;
{
    return(d_foChk(relNum, fldNum, valu, FTSTRING));
}
