#include "system.h"
#include "dbint.h"
#include "index.h"
#include "scan.h"
#ifdef MSDOS
#include "ctype.h"
#endif
#ifdef XENIX
#include <ctype.h>
#endif

extern int strcmp();

extern byte *d_alloc();
extern d_free();
extern d_abort();

extern XBLOCK *d_bxGet();
extern d_bgStuff();
extern d_bgStr();

extern d_rnFree();

extern bool d_xIns0();
extern bool d_xDel0();

/* these variables are referenced by other index routines through index.h: */

byte d_Key[MAXKEY];		/* actual key buffer for insert/delete */
long d_Tuple;			/* value being inserted/deleted */
RELATION *d_RelPtr;		/* relation pointer */
INDEX *d_Index;			/* index in use for insert/delete */
word d_KeyLen;			/* length of key for insert/delete */
word d_EntLen;			/* length of key + length of long */
word d_EntCnt;			/* number of entries/block */
bool d_Unique;			/* true if index insists on uniqueness */
bool d_Inode;			/* true if inserting into inode tree */
int (*d_Cmp)();			/* comparison routine for insert/delete */

/* these variables are only used in this file: */

SCANLIST *d_SFree;		/* list of inactive SCANLISTs */
POSSTACK *d_PFree;		/* list of inactive POSSTACKs */

SCANLIST *d_Scans;		/* list of active scans */
SCANLIST *d_Path;		/* path being built by d_xFnd1 */

SCANLIST *d_Sl;			/* scratch sl pointer */
POSSTACK *d_Ps;			/* scratch ps pointer */
int d_Pos;			/* scratch pos pointer */

word d_EOffs;			/* offset in XBLOCK of entries */
bool d_Pthng;			/* true if d_xFnd1 is pathing for d_xScn5 */

/*
 * d_xInit - indexing initialization routine.
 */

d_xInit()
{
    XBLOCK *b;

    d_Scans = nil;
    d_SFree = nil;
    d_PFree = nil;
    d_EOffs = (byte *)&b->x_entries[0] - (byte *)b;
}

/*
 * d_xTerm - indexing termination routine.
 */

#ifdef CLEANUP
d_xTerm()
{
    POSSTACK *pst;
    SCANLIST *slt;

    while (d_PFree != nil) {
	pst = d_PFree;
	d_PFree = d_PFree->ps_next;
	d_free((byte *)pst, sizeof(POSSTACK));
    }
    while (d_SFree != nil) {
	slt = d_SFree;
	d_SFree = d_SFree->sl_next;
	d_free((byte *)slt, sizeof(SCANLIST));
    }
}
#endif

/*
 * comparison routines used to compare key values.
 */

int
d_cmpBool(l, r)
    bool *l, *r;
{
    return(*l ? (*r ? 0 : 1) : (*r ? -1 : 0));
}

int
d_cmpShort(l, r)
    short *l, *r;
{
    return(*l < *r ? -1 : (*l == *r ? 0 : 1));
}

int
d_cmpLong(l, r)
    long *l, *r;
{
    return(*l < *r ? -1 : (*l == *r ? 0 : 1));
}

int
d_cmpFloat(l, r)
    double *l, *r;
{
    return(*l < *r ? -1 : (*l == *r ? 0 : 1));
}

int
d_cmpChar(l, r)
    byte *l, *r;
{
    register word len;

    len = d_KeyLen;
    while (len != 0 && *l == *r) {
	--len;
	l++;
	r++;
    }
    return(len == 0 ? 0 : *l - *r);
}

int
d_cmpStr(l, r)
    char *l, *r;
{
    extern int strncmp();

    return(strncmp(l, r, d_KeyLen - 1));
}

#ifdef CPM
		/* use faster routines (elsewhere) for MSDOS, XENIX */
int
d_cmpUC(l, r)
    char *l, *r;
{
    register word len;
    static char cl, cr;

    len = d_KeyLen - 1;
    do {
	cl = *l++;
	if (islower(cl)) {
	    cl = toupper(cl);
	}
	cr = *r++;
	if (islower(cr)) {
	    cr = toupper(cr);
	}
    } while (--len != 0 && cl != '\0' && cl == cr);
    return(cl - cr);
}

#else

int
d_cmpUC(l, r)
    char *l, *r;
{
    extern int strncmpuc();

    return(strncmpuc(l, r, d_KeyLen - 1));
}

#endif

/*
 * d_xSLen - setup d_KeyLen, d_EntLen and d_EntCnt from a given key length.
 */

d_xSLen(len)
    word len;
{

    d_KeyLen = len;
    d_EntLen = d_KeyLen + sizeof(TUPPTR) + sizeof(POINTER);
#ifdef M68K
    d_EntLen = (d_EntLen + 1) & ~1;	/* Round up to next multiple of 2 */
#endif
    d_EntCnt = ((d_Inode ? INOSIZE : BLKSIZE) - d_EOffs) / d_EntLen;
}

/*
 * d_xSetup - routine to set up specifics for insert, delete and lookup.
 */

d_xSetup(relPtr, x)
    RELATION *relPtr;
    register INDEX *x;
{

    d_Inode = false;
    d_RelPtr = relPtr;
    d_Index = x;
    d_xSLen(x->x_len);
    d_Unique = x->x_unique;
    switch (x->x_type) {
    case FTBOOL:
	d_Cmp = d_cmpBool;
	break;
    case FTSHORT:
	d_Cmp = d_cmpShort;
	break;
    case FTLONG:
	d_Cmp = d_cmpLong;
	break;
    case FTFLOAT:
	d_Cmp = d_cmpFloat;
	break;
    case FTCHAR:
	d_Cmp = d_cmpChar;
	break;
    case FTSTRING:
	d_Cmp = d_cmpStr;
	break;
    case FTUCSTR:
	d_Cmp = d_cmpUC;
	break;
    default:
	d_abort(18);
    }
}

/*
 * d_xpAlloc - allocate a POSSTACK.
 *	       Can't use d_Ps as a temporary here since this routine is
 *	       called (from d_xIns0) when a value is kept there.
 */

POSSTACK *
d_xpAlloc()
{
    register POSSTACK *ps;

    if (d_PFree != nil) {
	ps = d_PFree;
	d_PFree = d_PFree->ps_next;
    } else {
	ps = (POSSTACK *)d_alloc(sizeof(POSSTACK));
    }
    return(ps);
}

/*
 * d_xpFree - free a POSSTACK.
 */

d_xpFree(ps)
    register POSSTACK *ps;
{
    ps->ps_next = d_PFree;
    d_PFree = ps;
}

/*
 * d_xsAlloc - allocate a SCANLIST.
 */

SCANLIST *
d_xsAlloc()
{
    register SCANLIST *sl;

    if (d_SFree != nil) {
	sl = d_SFree;
	d_SFree = d_SFree->sl_next;
    } else {
	sl = (SCANLIST *)d_alloc(sizeof(SCANLIST));
    }
    return(sl);
}

/*
 * d_xsFree - free a SCANLIST.
 */

d_xsFree(sl)
    register SCANLIST *sl;
{
    sl->sl_next = d_SFree;
    d_SFree = sl;
}

/*
 * d_xFnd1 - lower, recursive level of B-tree lookup.
 */

bool
d_xFnd1(key, node, valu)
    byte *key;
    POINTER *node;
    long *valu;
{
    register XBLOCK *n;
    static ENTRY *pLow;
    POINTER deeper;			/* cannot be 'static' */
    static word count, high, mid;
    static bool found;

    found = false;
    n = d_bxGet(d_RelPtr, node);

    if ((*d_Cmp)(key, &n->x_entries[0].e_key[0]) < 0) {
	d_cpPtr(&deeper, &n->x_left);
	d_Pos = -1;
    } else {

	/* perform a binary search to find the appropriate value: */

	count = n->x_count;

	d_Pos = 0;
	high = count + 1;
	while (d_Pos + 1 != high) {
	    mid = (d_Pos + high) / 2;
	    if ((*d_Cmp)(key,
			 &n->x_entries[0].e_key[0] + (mid-1)*d_EntLen) > 0) {
		d_Pos = mid;
	    } else {
		high = mid;
	    }
	}

	pLow = (ENTRY *)((byte *)&n->x_entries[0] + d_Pos * d_EntLen);
	if (d_Pos != count && (*d_Cmp)(key, &pLow->e_key[0]) == 0) {
	    found = true;
	    *valu = pLow->e_tuple.t_long;
	} else {
	    d_cpPtr(&deeper, &((ENTRY *)((byte *)pLow - d_EntLen))->e_down);
	    /* if not at bottom, want to point AT the one we go down under */
	    if (!n->x_isBottom) {
		--d_Pos;
	    }
	}
    }

    if (d_Pthng) {
	/* A scan path will be set up which points at the found key, or
	   which points at the bottom level slot which the unfound key would
	   have been directly before. If ps_pos == -1, then the unfound key
	   would have been before the first slot in the given node (which
	   means that it is either before all keys in the tree or that it
	   comes directly after the slot in the immediate parent node which
	   points at this node). If the key wasn't found, then the scan will
	   always go all the way down to the bottom of the tree. */
	d_Path->sl_depth++;
	d_Ps = d_xpAlloc();
	d_Ps->ps_next = d_Path->sl_pos;
	d_cpPtr(&d_Ps->ps_ptr, node);
	d_Ps->ps_pos = d_Pos;
	d_Path->sl_pos = d_Ps;
    }

    return((found || n->x_isBottom) ? found : d_xFnd1(key, &deeper, valu));
}

/*
 * d_xFnd0 - top level of index lookup routine.
 */

long
d_xFnd0(key, root)
    byte *key;
    register POINTER *root;
{
    static long valu;

    return((root->p_file == 0 || !d_xFnd1(key, root, &valu)) ? 0L : valu);
}

/*
 * d_xFPath - free a POSSTACK stack.
 */

d_xFPath(ps)
register POSSTACK *ps;
{
    static POSSTACK *temp;

    while (ps != nil) {
	temp = ps;
	ps = ps->ps_next;
	d_xpFree(temp);
    }
}

/*
 * d_xMod - an index is being modified - reset scans.
 */

d_xMod()
{
    register SCANLIST *sl;

    sl = d_Scans;
    while (sl != nil) {
	if (sl->sl_index == d_Index && sl->sl_state == SL_NORMAL) {
	    sl->sl_state = SL_REFIND;
	    d_xFPath(sl->sl_pos);
	}
	sl = sl->sl_next;
    }
}

/*
 * d_xSet - common setup for d_xIns and d_xDel.
 */

d_xSet(tuple)
    long tuple;
{
    static int (*getter)();
    static long zero;

    zero = 0L;
    d_Tuple = tuple;
    if (d_Tuple <= zero) {
	d_abort(22);
    }
    getter = ((d_Index->x_type == FTSTRING || d_Index->x_type == FTUCSTR) ?
	      d_bgStr : d_bgStuff);
    (*getter)(d_RelPtr,
	      (d_Tuple - 1L) * d_RelPtr->r_len +
		d_RelPtr->r_fields[d_Index->x_field].rf_pos,
	      &d_Key[0], d_KeyLen);
    if (d_RelPtr->r_fields[d_Index->x_field].rf_scnCnt != 0) {
	d_xMod();
    }
}

/*
 * d_xIns - insert a tuple into the already setup index.
 */

d_xIns(tuple)
    long tuple;
{
    d_xSet(tuple);
    d_xIns0(&d_Index->x_root);
}

/*
 * d_xDel - delete a tuple from an already setup index.
 */

d_xDel(tuple)
    long tuple;
{
    d_xSet(tuple);
    d_xDel0(&d_Index->x_root);
}

/*
 * d_xInsV - insert a tuple into an already setup virtual index.
 */

d_xInsV(tuple, key)
    long tuple;
    byte *key;
{
    COPY(&d_Key[0], key, d_KeyLen);
    d_Tuple = tuple;
    d_xMod();
    d_xIns0(&d_Index->x_root);
}

/*
 * d_xDelV - delete a tuple from an already setup virtual index.
 */

d_xDelV(tuple, key)
    long tuple;
    byte *key;
{
    COPY(&d_Key[0], key, d_KeyLen);
    d_Tuple = tuple;
    d_xMod();
    d_xDel0(&d_Index->x_root);
}

/*
 * d_xFnd - look up a value in a B-tree (and possibly it's inode).
 */

long
d_xFnd(key)
    byte *key;
{
    static TUPPTR res;

    d_Pthng = false;
    res.t_long = d_xFnd0(key, &d_Index->x_root);
    if (!d_Unique && res.t_long < 0L) {
	if (res.t_ptr.p_file == 0) {
	    return(0L);
	} else {
	    res.t_long =
		d_bxGet(d_RelPtr, &res.t_ptr)->x_entries[0].e_tuple.t_long;
	}
    }
    return(res.t_long);
}

/*
 * d_nSet - common code to set up for d_nIns, d_nDel or d_xScn4.
 */

d_nSet()
{

    d_Inode = true;
    d_xSLen(sizeof(long));
    d_Unique = true;
    d_Cmp = d_cmpLong;
}

/*
 * d_nIns - insert a tuple into the pointed-to inode tree.
 *	    d_nIns is called from within d_xIns1, which is called from
 *	    d_xIns0. The double recursion will only happen once, however,
 *	    and if d_xIns1 calls d_nIns, it is done, hence the values it
 *	    was using need not be restored here. We return 'true' if the
 *	    value pointed to by 'ptr' has been changed (either directly here
 *	    or by d_xIns0 or d_xIns1). Assumption: the space pointed to by
 *	    the parameter will not be invalidated by calls to d_xIns0.
 */

bool
d_nIns(ptr)
    register TUPPTR *ptr;
{
    static long temp;
    static bool newTree;

    d_nSet();
    newTree = false;
    temp = 0L;
    if ((*ptr).t_long >= temp) {
	newTree = true;
	temp = d_Tuple;
	d_Tuple = (*ptr).t_long;
	(*ptr).t_ptr.p_file = 0;	/* null ptr for d_xIns0 to grow */
	COPY(&d_Key[0], &d_Tuple, sizeof(long));
	d_xIns0(&(*ptr).t_ptr);
	d_Tuple = temp;
    }
    COPY(&d_Key[0], &d_Tuple, sizeof(long));
    if (d_xIns0(&(*ptr).t_ptr)) {
	newTree = true;
    }
    return(newTree);
}

/*
 * d_nDel - delete a tuple from an inode tree (if it's a tree).
 *	    d_xDel0 tells us if it changed the root, but does not tell us
 *	    if the tree has degenerated into a single entry, so we must
 *	    access the root block and check the count. If the count is 1,
 *	    then we throw away the whole tree and replace it by the single
 *	    remaining tuple. We return 'true' if the root pointer
 *	    has been changed. Note that this routine MUST restore the index
 *	    environment when it returns. Note also that if the root pointer
 *	    has been changed to a null pointer, then the entry must be
 *	    deleted from the main index B-tree. Assumption: the space pointed
 *	    to by the parameter will not be invalidated by calls to d_xDel0.
 */

bool
d_nDel(tupPtr)
    TUPPTR *tupPtr;
{
    register XBLOCK *n;
    static long temp;
    static bool res;

    d_nSet();
    COPY(&d_Key[0], &d_Tuple, sizeof(long));
    res = d_xDel0(tupPtr);
    n = d_bxGet(d_RelPtr, &(*tupPtr).t_ptr);
    if (n->x_isBottom && n->x_count == 1) {
	temp = n->x_entries[0].e_tuple.t_long;
	d_rnFree(d_RelPtr, &(*tupPtr).t_ptr);
	(*tupPtr).t_long = temp;
	res = true;
    }
    d_xSetup(d_RelPtr, d_Index);
    return(res);
}

/*
 * d_xScnx - low-level common code for setting up a scan.
 */

SCANLIST *
d_xScnx(r, x)
    RELATION *r;
    register INDEX *x;
{
    d_Sl = d_xsAlloc();
    d_Sl->sl_next = d_Scans;
    d_Scans = d_Sl;
    d_Sl->sl_rel = r;
    d_Sl->sl_index = x;
    d_Sl->sl_depth = 0;
    d_Sl->sl_inode = false;
    if (x->x_field >= 0) {
	r->r_fields[x->x_field].rf_scnCnt += 1;
    }
    return(d_Sl);
}

/*
 * d_xScn1 - setup a scan of an index.
 */

SCANLIST *
d_xScn1(r, x, leftToRight)
    RELATION *r;
    register INDEX *x;
    bool leftToRight;
{
    if (x->x_root.p_file == 0) {
	return(nil);
    } else {
	d_Sl = d_xScnx(r, x);
	d_Sl->sl_state = leftToRight ? SL_LEFT : SL_RIGHT;
	return(d_Sl);
    }
}

/*
 * d_xScn4 - setup a scan of an inode structure.
 */

SCANLIST *
d_xScn4(r, x, key, leftToRight)
    RELATION *r;
    register INDEX *x;
    byte *key;
    bool leftToRight;
{
    d_Sl = d_xScnx(r, x);
    d_Sl->sl_inode = true;
    COPY(&d_Sl->sl_key[0], key, x->x_len);
    d_Sl->sl_state = leftToRight ? SL_LEFT : SL_RIGHT;
    return(d_Sl);
}

/*
 * d_xScn5 - setup a scan of an index tree at a specific point.
 */

SCANLIST *
d_xScn5(r, x, key)
    RELATION *r;
    register INDEX *x;
    byte *key;
{

    if (x->x_root.p_file == 0) {
	return(nil);
    } else {
	d_Sl = d_xScnx(r, x);
	COPY(&d_Sl->sl_key[0], key, x->x_len);
	d_Sl->sl_state = SL_TARGET;
	return(d_Sl);
    }
}

/*
 * d_sDown - scan down a tree to a bottom node.
 *	     If 'leftToRight' is true, then scan down leftwards (towards the
 *	     lexically least keys), else scan down rightwards.
 *	     ASSUMPTION: d_EntLen has been set up.
 */

POSSTACK *
d_sDown(sl, oldps, ptr, leftToRight)
    SCANLIST *sl;
    POSSTACK *oldps;
    register POINTER *ptr;
    bool leftToRight;
{
    static XBLOCK *b;

    do {
	sl->sl_depth++;
	d_Ps = d_xpAlloc();
	d_Ps->ps_next = oldps;
	oldps = d_Ps;
	d_cpPtr(&d_Ps->ps_ptr, ptr);
	b = d_bxGet(sl->sl_rel, ptr);
	if (leftToRight) {
	    d_Ps->ps_pos = -1;		/* -1 => at the left slot */
	    ptr = &b->x_left;
	} else {
	    d_Ps->ps_pos = b->x_count - 1;
	    ptr = &((ENTRY *)((byte *)&b->x_entries[0] +
			      d_EntLen * d_Ps->ps_pos))->e_down;
	}
    } while (!b->x_isBottom);
    if (leftToRight) {
	d_Ps->ps_pos = 0;
    }
    return(d_Ps);
}

/*
 * d_xScnN - SL_NORMAL scan - just step along.
 */

word
d_xScnN(sl, pKey, pVal, leftToRight)
    register SCANLIST *sl;
    byte *pKey;
    long *pVal;
    bool leftToRight;
{
    static XBLOCK *b;
    static POSSTACK *temp;
    static ENTRY *e;

    d_Inode = sl->sl_inode;
    d_xSLen(d_Inode ? sizeof(long) : sl->sl_index->x_len);
    d_Ps = sl->sl_pos;
    d_Pos = d_Ps->ps_pos;
    b = d_bxGet(sl->sl_rel, &d_Ps->ps_ptr);
    if (leftToRight) {
	if (b->x_isBottom) {
	    /* try to move forward; climbing tree if needed: */
	    while (true) {
		d_Pos += 1;
	    if (d_Pos != b->x_count) break;
		temp = d_Ps;
		d_Ps = d_Ps->ps_next;
		d_xpFree(temp);
		sl->sl_pos = d_Ps;
		if (d_Ps == nil) {
		    return(0);
		}
		d_Pos = d_Ps->ps_pos;
		b = d_bxGet(sl->sl_rel, &d_Ps->ps_ptr);
		sl->sl_depth--;
	    }
	    d_Ps->ps_pos = d_Pos;
	} else {
	    /* not at bottom - go down from our current position: */
	    d_Ps = d_sDown(sl, d_Ps, &((ENTRY *)((byte *)&b->x_entries[0] +
					         d_EntLen * d_Pos))->e_down,
			   true);
	    b = d_bxGet(sl->sl_rel, &d_Ps->ps_ptr);
	    d_Pos = 0;
	}
    } else {
	d_Pos -= 1;
	d_Ps->ps_pos = d_Pos;
	if (b->x_isBottom) {
	    /* on a bottom node - must go up until we can go left,
	       then yield that slot. */
	    while (d_Pos == -1) {
		temp = d_Ps;
		d_Ps = d_Ps->ps_next;
		d_xpFree(temp);
		sl->sl_pos = d_Ps;
		if (d_Ps == nil) {
		    return(0);
		}
		d_Pos = d_Ps->ps_pos;
		b = d_bxGet(sl->sl_rel, &d_Ps->ps_ptr);
		sl->sl_depth--;
	    }
	} else {
	    d_Ps =
		d_sDown(sl, d_Ps,
			(d_Pos == -1 ?
			     &b->x_left
			 :
			     &((ENTRY *)((byte *)&b->x_entries[0] +
					 d_EntLen * d_Pos))->e_down),
			false);
	    b = d_bxGet(sl->sl_rel, &d_Ps->ps_ptr);
	    d_Pos = d_Ps->ps_pos;
	}
    }
    e = (ENTRY *)((byte *)&b->x_entries[0] + d_EntLen * d_Pos);
    if (d_Inode) {
	COPY((byte *) &sl->sl_tuple, &e->e_key[0], d_KeyLen);
    } else {
	COPY(&sl->sl_key[0], &e->e_key[0], d_KeyLen);
    }
    if (pKey != nil) {
	COPY(pKey, &e->e_key[0], d_KeyLen);
    }
    *pVal = e->e_tuple.t_long;
    sl->sl_pos = d_Ps;
    return(sl->sl_depth);
}

/*
 * d_xScn2 - start/continue a scan of an index.
 */

word
d_xScn2(sl, pKey, pVal, leftToRight)
    register SCANLIST *sl;
    byte *pKey;
    long *pVal;
    bool leftToRight;
{
    static ENTRY *e;
    static TUPPTR root;
    static long zero;
    static bool move;

    /* do a quick check for the normal case: */

    if (sl->sl_state == SL_NORMAL) {
	return(d_xScnN(sl, pKey, pVal, leftToRight));
    }

    if (sl->sl_state == SL_GONE ||
	    sl->sl_state == SL_LEFT && !leftToRight ||
	    sl->sl_state == SL_RIGHT && leftToRight ||
	    sl->sl_index->x_root.p_file == 0) {
	return(0);
    }

    d_xSetup(sl->sl_rel, sl->sl_index);

    if (sl->sl_inode) {
	d_Pthng = false;
	root.t_long = d_xFnd0(&sl->sl_key[0], &sl->sl_index->x_root);
	zero = 0L;
	if (root.t_long == 0L) {
	    /* no longer any tuples with this key in the tree: */
	    return(0);
	}
	if (root.t_long > zero) {
	    /* only one tuple with this key left now - just yield it: */
	    sl->sl_state = SL_GONE;
	    *pVal = root.t_long;
	    /* inode scan NEVER has a pKey other than nil. */
	    return(1);
	}
	d_nSet();
    } else {
	d_Inode = false;
	d_cpPtr(&root.t_ptr, &sl->sl_index->x_root);
    }

    sl->sl_depth = 0;
    if (sl->sl_state == SL_LEFT || sl->sl_state == SL_RIGHT) {
	sl->sl_pos = d_sDown(sl, nil, &root.t_ptr, leftToRight);
	move = false;
    } else {
	sl->sl_pos = nil;
	d_Path = sl;
	d_Pthng = true;
	root.t_long =
	    d_xFnd0(
		(d_Inode ?
		     (byte *) &sl->sl_tuple
		 :
		     &sl->sl_key[0]),
		&root.t_ptr
	    );
	if (root.t_long == 0L) {
	    /* we didn't find the one we were after. Thus, we are now
	       pointing where it should have been. If we are going
	       leftToRight, then we have to watch out for the fact that the
	       one we wanted might have been just after the last entry in 
	       this node. We handle that by backing up and then MOVEing. If
	       we are going RightToLeft, we always want to move back up. */
	    if (sl->sl_pos->ps_pos == -1) {
		sl->sl_pos->ps_pos = 0;
	    }
	    if (leftToRight) {
		if (sl->sl_pos->ps_pos == 0) {
		    move = false;
		} else {
		    sl->sl_pos->ps_pos -= 1;
		    move = true;
		}
	    } else {
		move = true;
	    }
	} else {
	    /* we found the one we were searching for. If we were just
	       starting the scan then we are all set, otherwise we move on
	       from there. */
	    move = sl->sl_state == SL_REFIND;
	}
    }

    sl->sl_state = SL_NORMAL;

    if (move) {
	return(d_xScnN(sl, pKey, pVal, leftToRight));
    }

    /* we are pointing at the slot we want. Access it and return things. */

    d_Ps = sl->sl_pos;
    e = (ENTRY *)((byte *)&d_bxGet(d_RelPtr, &d_Ps->ps_ptr)->x_entries[0] +
		  d_EntLen * d_Ps->ps_pos);
    if (d_Inode) {
	COPY((byte *) &sl->sl_tuple, &e->e_key[0], d_KeyLen);
    } else {
	COPY(&sl->sl_key[0], &e->e_key[0], d_KeyLen);
    }
    if (pKey != nil) {
	COPY(pKey, &e->e_key[0], d_KeyLen);
    }
    *pVal = e->e_tuple.t_long;
    return(sl->sl_depth);
}

/*
 * d_xScn3 - release a scan (terminated or otherwise).
 */

d_xScn3(sl)
    SCANLIST *sl;
{
    register SCANLIST **slp;
    static int fldNum;

    if (sl->sl_state == SL_NORMAL) {
	d_xFPath(sl->sl_pos);
    }
    slp = &d_Scans;
    while (*slp != nil && *slp != sl) {
	slp = &(*slp)->sl_next;
    }
    if (*slp == nil) {
	d_abort(17);
    }
    *slp = sl->sl_next;
    fldNum = sl->sl_index->x_field;
    if (fldNum >= 0) {
	sl->sl_rel->r_fields[fldNum].rf_scnCnt -= 1;
    }
    d_xsFree(sl);
}
