#include "system.h"
#include "dbint.h"
#include "index.h"

extern d_cpPtr();

extern XBLOCK *d_bxGet();
extern d_bxMark();

extern d_rxAlloc();
extern d_rnAlloc();

extern bool d_nIns();

/*
 * d_xIns1 - low-level (recursive) routine to insert a key into an index.
 *	     Returns 'true' if the node at this level was split, and hence
 *	     the higher-level call must handle the node passed back. In that
 *	     case, the key value to use is passed back in d_Key, a POINTER
 *	     to the node to insert is passed back in 'back' and the TUPPTR
 *	     for that slot is passed back in 'bTuple'.
 */

bool
d_xIns1(node, back, bTuple)
    POINTER *node, *back;
    TUPPTR *bTuple;
{
    XBLOCK *n, *n1, *newN;		/* pointers to blocks */
    ENTRY *ep1, *ep2, *backEnt;		/* pointers to key-value pairs */
    POINTER node2, newNode, thisNode;	/* pointers to nodes */
    TUPPTR inode;
    word mid, end;
    word count, i;
    bool isBottom;

    n = d_bxGet(d_RelPtr, node);
    isBottom = n->x_isBottom;
    count = n->x_count;

    i = 0;
    ep1 = &n->x_entries[0];

    if ((*d_Cmp)(&d_Key[0], &ep1->e_key[0]) < 0) {

	/* new key is less than first - need to insert before first: */

	d_cpPtr(&node2, &n->x_left);

    } else {

	/* perform a binary search to find the appropriate slot: */

	end = count + 1;
	while (i + 1 != end) {
	    mid = (i + end) / 2;
	    if ((*d_Cmp)(&d_Key[0], &ep1->e_key[0]+(mid-1)*d_EntLen) > 0) {
		i = mid;
	    } else {
		end = mid;
	    }
	}

	ep1 = (ENTRY *)((byte *)&n->x_entries[0] + i * d_EntLen);

	/* at this point, i and ep1 either point at the slot where our key is
	   found, or, our key isn't in this node and i and ep1 point at the
	   slot our key should be inserted before. Note that this could mean
	   that they point past the end of the node. */

	if (i != count && (*d_Cmp)(&d_Key[0], &ep1->e_key[0]) == 0) {

	    /* an exact match with the key (entry already existed): */

	    if (d_Unique) {
		ep1->e_tuple.t_long = d_Tuple;
		d_bxMark(n, node);
	    } else {

		/* need temporary copy of the pointer, since inserting 
		   into the inode tree may make our buffer go away. */

		inode.t_long = ep1->e_tuple.t_long;
		end = d_EntLen;		/* save, since d_nIns changes it */
		if (d_nIns(&inode)) {

		    /* d_nIns (which calls us recursively to insert into an
		       inode tree) has yielded a new root to the inode tree.
		       We must put that root into the slot here, and mark
		       this block as dirty. */

		    n = d_bxGet(d_RelPtr, node);
		    ((ENTRY *)((byte *)&n->x_entries[0] + i * end))->
			    e_tuple.t_long = inode.t_long;
		    d_bxMark(n, node);
		}
	    }
	    return(false);
	}

	d_cpPtr(&node2, &((ENTRY *)((byte *)ep1 - d_EntLen))->e_down);
    }

    /* At this point 'node2' contains the POINTER down to a deeper node that
       we must recurse to. We don't recurse if we are already at the bottom
       of the tree, but we still have to do the other stuff. */

    if (!isBottom && !d_xIns1(&node2, back, bTuple)) {

	/* no more work to be done: */

	return(false);

    } else {

	/* insert new entry here */

	/* recursive call may have re-used buffer block!! */

	if (!isBottom) {
	    n = d_bxGet(d_RelPtr, node);
	}
	n1 = n;
	d_cpPtr(&thisNode, node);
	if (count == d_EntCnt) {

	    /* no room here, split this node and pass on up: */

	    mid = d_EntCnt / 2;		/* note that we round down */

	    /* entries 0 -> 'mid'-1 stay in the existing node. Entry 'mid'
	       becomes the LEFT entry of the new node. (It's key and 'tuple'
	       parts go up to the new entry inserted into our parent.)
	       Entries 'mid'+1 -> d_EntCnt - 1 become entries in new node.

	       NOTE: IT IS FOR THIS REASON THAT THE MINIMUM NUMBER OF SLOTS
	       IN AN INDEX TREE NODE IS 3. */

	    (d_Inode ? d_rnAlloc : d_rxAlloc)(d_RelPtr, &newNode);
	    d_cpPtr(&node2, &newNode);
	    newN = d_bxGet(d_RelPtr, &newNode);
	    newN->x_isBottom = n->x_isBottom;
	    backEnt = (ENTRY *)((byte *)&n->x_entries[0] + mid * d_EntLen);
	    d_cpPtr(&newN->x_left, &backEnt->e_down);
	    end = d_EntCnt - mid - 1;
	    COPY(&newN->x_entries[0],(byte *)backEnt+d_EntLen, end*d_EntLen);
	    n->x_count = mid;
	    newN->x_count = end;
	    if (i > mid) {
		n = newN;
		d_cpPtr(&thisNode, &newNode);
		i = i - mid - 1;
		mid = end;
	    }

	    /* A bit of a cheat here. We don't want to have a second MAXKEY
	       length buffer to hold the key that is sent back to our parent.
	       We can't copy it into d_Key right away, since we haven't moved
	       out the key that is in there yet. We can't just leave it as is
	       since we will be overwriting it. Hence we move it to a slot
	       in the original node that we know is beyond reach of the
	       shuffle, and make 'backEnt' point there. */

	    ep1 = (ENTRY *)((byte *)&n->x_entries[0] + (mid+1) * d_EntLen);
	    COPY(ep1, backEnt, d_EntLen);
	    backEnt = ep1;
	} else {

	    /* room here, just mark the end of the shuffle: */

	    mid = count;
	}

	/* Shuffle entries right to make room for the new one. 'i' is the
	   index of where the new element is to be inserted; 'end' is index
	   of just past the last slot to be shuffled (i.e. is the slot that
	   will be newly filled by the shuffle). */

	ep1 = (ENTRY *)((byte *)&n->x_entries[0] + mid * d_EntLen);
	while (mid-- != i) {
	    ep2 = (ENTRY *)((byte *)ep1 - d_EntLen);
	    COPY(ep1, ep2, d_EntLen);
	    ep1 = ep2;
	}

	/* Insert new entry at its proper slot. After the shuffle, 'ep1' just
	   happens to point at the slot to be used. Note the subtlety here:
	   the key to insert, whether from a simple insertion here at the
	   bottom, or from a key passed up from a lower call that split it's
	   node, the key value is in d_Key. In the first case, we have no
	   'down' pointer to bother with and the tuple value is just the one
	   being inserted into the tree. In the second case, we have a new
	   'down' pointer and the tuple comes up from the recursion. */

	COPY(&ep1->e_key[0], &d_Key[0], d_KeyLen);
	if (isBottom) {
	    ep1->e_tuple.t_long = d_Tuple;
	} else {
	    ep1->e_tuple.t_long = (*bTuple).t_long;
	    d_cpPtr(&ep1->e_down, back);
	}
	n->x_count++;

	/* we have changed this node, flag it as dirty: */

	d_bxMark(n1, node);

	/* if we split, flag the new node, and pass info on up: */

	if (count == d_EntCnt) {
	    d_bxMark(newN, &newNode);
	    d_cpPtr(back, &node2);
	    (*bTuple).t_long = backEnt->e_tuple.t_long;
	    COPY(&d_Key[0], &backEnt->e_key[0], d_KeyLen);
	    return(true);
	} else {
	    return(false);
	}
    }
}

/*
 * d_xIns0 - top level routine to insert a key into an index.
 *	     Passed are a pointer to the place which holds the root pointer
 *	     (it may get updated) and the tupleId to insert into the tree.
 *	     'true' is returned if the root pointer is changed.
 */

d_xIns0(root)
    POINTER *root;
{
    POINTER back, newRoot;
    TUPPTR tupPtr;
    XBLOCK *n;

    if (root->p_file == 0) {

	/* nothing in tree yet - create a root node and insert into it: */

	(d_Inode ? d_rnAlloc : d_rxAlloc)(d_RelPtr, root);
	n = d_bxGet(d_RelPtr, root);
	n->x_isBottom = true;
	n->x_count = 1;
	n->x_left.p_file = 0;		/* needed so d_xDel gets null */
	COPY(&n->x_entries[0].e_key[0], &d_Key[0], d_KeyLen);
	n->x_entries[0].e_tuple.t_long = d_Tuple;
	d_bxMark(n, root);
	return(true);

    } else {

	/* tree exists, insert into it: */

	if (d_xIns1(root, &back, &tupPtr)) {

	    /* have to split the root node into two halves, the old
	       root being the left half, and the returned value being the
	       right half (2nd of two entries in the new root).

	       Note that if we inserted into an inode tree and then returned
	       from the top level call of d_xIns1, we won't ever get here
	       since if we inserted into an inode tree, the key WAS already
	       in the top-level tree. */

	    (d_Inode ? d_rnAlloc : d_rxAlloc)(d_RelPtr, &newRoot);
	    n = d_bxGet(d_RelPtr, &newRoot);
	    n->x_isBottom = false;
	    n->x_count = 1;
	    d_cpPtr(&n->x_left, root);
	    COPY(&n->x_entries[0].e_key[0], &d_Key[0], d_KeyLen);
	    n->x_entries[0].e_tuple.t_long = tupPtr.t_long;
	    d_cpPtr(&n->x_entries[0].e_down, &back);
	    d_bxMark(n, &newRoot);
	    d_cpPtr(root, &newRoot);
	    return(true);
	} else {
	    return(false);
	}
    }
}
