#include "system.h"
#include "dbint.h"
#include "index.h"

extern d_cpPtr();
extern bool d_ptrEq();

extern XBLOCK *d_bxGet();
extern d_bxMark();

extern d_rxFree();
extern d_rnFree();

extern bool d_nDel();

TUPPTR d_TupPtr;		/* 'tuple' from bottom to deleted slot */

/*
 * d_xGone - routine to call if deletion cannot find tuple to delete.
 */

d_xGone()
{
}

/*
 * d_xDel1 - lower-level, recursive routine to delete a key-value from the
 *	     B-tree.
 *	     On the way down the tree during recursion, 'd_Key' is the key
 *	     to be deleted from the tree. 'node' is the node to look in this
 *	     time and 'found' is true if the key has been found in an upper
 *	     node and we are just looking for a value to replace it (that
 *	     value will be the leftmost key in the bottom-left descendant
 *	     of the slot in which we found the key to delete).
 *	     On the way back up the tree, when the recursive calls are
 *	     returning, the returned 'bool' is true if the node we are moving
 *	     up from has no keys left and hence must be either deleted or
 *	     must get a key from a neighbour (a sibling). If the result is
 *	     true, *pp will be set to the pointer to the node that has been
 *	     emptied and that node will still be in a buffer and will not
 *	     have been marked as modified yet. The replacement key value
 *	     will be in 'd_Key' on the way back up.
 */

bool
d_xDel1(node, found, pp)
    POINTER *node;
    bool found;
    XBLOCK **pp;
{
    register ENTRY *ep1;
    XBLOCK *n, *pLeft, *pRight;	/* pointers to nodes */
    ENTRY *ep2;			/* pointers to entries within nodes */
    POINTER node2, nLeft, nRight;
    TUPPTR tupPtr;		/* used when calling d_nDel */
    word count, i, j, k;
    bool atLeft, weFound, emptyChild;

    weFound = false;
    n = d_bxGet(d_RelPtr, node);
    count = n->x_count;

    i = 0;
    ep1 = &n->x_entries[0];
    if ((*d_Cmp)(&d_Key[0], &ep1->e_key[0]) < 0) {

	/* key is less than first - need to look before first: */

	d_cpPtr(&node2, &n->x_left);
	atLeft = true;

    } else {

	/* perform a binary search to find the appropriate slot: */

	j = count + 1;
	while (i + 1 != j) {
	    k = (i + j) / 2;
	    if ((*d_Cmp)(&d_Key[0], &ep1->e_key[0] + (k-1) * d_EntLen) > 0) {
		i = k;
	    } else {
		j = k;
	    }
	}

	ep1 = (ENTRY *)((byte *)ep1 + i * d_EntLen);
	if (i != count && (*d_Cmp)(&d_Key[0], &ep1->e_key[0]) == 0) {

	    /* an exact match with the key (entry existed): */

	    ep2 = ep1;
	    ep1 = (ENTRY *)((byte *)ep1 + d_EntLen);
	    i++;
	    weFound = true;
	    found = true;

	    /* we found the key at this level - we want to
	       delete it, moving those after it one slot to the left. If
	       this is a non-unique index, however, we will only do that if
	       the entry isn't an inode tree. If it is an inode tree, we
	       call d_nDel to delete from it. If it reports that it changed
	       the root of that tree, then we put the new root in our slot
	       and mark it as changed */

	    if (!d_Unique && ep2->e_tuple.t_long < 0L) {
		tupPtr.t_long = ep2->e_tuple.t_long;
		if (d_nDel(&tupPtr.t_ptr)) {
		    n = d_bxGet(d_RelPtr, node);
		    ep1 = (ENTRY *)((byte *)&n->x_entries[0] + i * d_EntLen);
		    ep2 = (ENTRY *)((byte *)ep1 - d_EntLen);
		    ep2->e_tuple.t_long = tupPtr.t_long;
		    d_bxMark(n, node);
		}
		/* in either case, we are all done - there will be no
		   modification to the main tree. */
		return(false);
	    }
	} else {
	    ep2 = (ENTRY *)((byte *)ep1 - d_EntLen);
	}
	d_cpPtr(&node2, &ep2->e_down);
	atLeft = false;
    }

    /* At this point: ep1 points at slot i; if i == 0 then atLeft is true
       and the searched for key is less than any in this node. If the key
       was found in this node, then i and ep1 point to the slot after it
       and ep2 points at that slot.
       If the key wasn't found, then i and ep1 point to the node after where
       it should have been and ep2 points at the one before.

       If we got this far (i.e. we didn't just delete from an inode), we must
       always go to the very bottom of the tree: */

    if (n->x_isBottom) {
	if (!weFound) {
	    if (found) {

		/* key was found higher up, but the first slot in this bottom
		   node is the leftmost under the deleted one. It is from
		   here that we get a key-tuple pair to replace the deleted
		   one. This can result in all sorts of tree collapsing.
		   Shift bottom node left and set 'd_Key' and 'd_TupPtr'.
		   Because of structure of tree, atLeft is true here. We also
		   setup up i, ep1, and ep2 for the shift of the node. */

		COPY(&d_Key[0], &ep1->e_key[0], d_KeyLen);
		d_TupPtr.t_long = ep1->e_tuple.t_long;
		i = 1;
		ep2 = ep1;
		ep1 = (ENTRY *)((byte *)ep1 + d_EntLen);
	    } else {

		/* EEEEK - the key we were searching for to delete isn't
		   in the tree at all. We call d_xGone (who can do something
		   if he wants) and then just return. */

		d_xGone();
		return(false);
	    }
	}

	/* shift entries left to fill gap left by deletion: */

	while (i++ != count) {
	    COPY(ep2, ep1, d_EntLen);
	    ep2 = ep1;
	    ep1 = (ENTRY *)((byte *)ep1 + d_EntLen);
	}
	count--;
	n->x_count = count;
	if (count == 0) {

	    /* this node has no entry - pass indication back up:
	       (note that we re-get this node so that the LRU block
	       handling code will put it back as the most recent, so
	       that it will be available after the next level up of
	       the recursion has gotten two other nodes (including
	       the main node for that level). */

	    *pp = d_bxGet(d_RelPtr, node);
	    return(true);
	} else {
	    d_bxMark(n, node);
	    return(false);
	}

    }

    /* we have to recurse down. When we get back, our node may not be
       still in it's buffer, so we make our pointers relative, and then
       reproduce them after the recursion, at which time they may be
       pointing into a different buffer. */

    emptyChild = d_xDel1(&node2, found, pp);

    n = d_bxGet(d_RelPtr, node);
    ep1 = (ENTRY *)((byte *)&n->x_entries[0] + i * d_EntLen);
    ep2 = (ENTRY *)((byte *)ep1 - d_EntLen);

    /* if the key was in this node, then the lower calls would have
       assigned a replacement value into 'd_Key' and 'd_TupPtr'. Put
       that value in, before doing anything else: */

    if (weFound) {
	COPY(&ep2->e_key[0], &d_Key[0], d_KeyLen);
	ep2->e_tuple.t_long = d_TupPtr.t_long;
	d_bxMark(n, node);
    }

    if (!emptyChild) {

	/* the tree below us didn't collapse - we are all done */

	return(false);
    }

    /* child node has no entries left. Check for a sibling who can
       share one of it's entries: */

    if (i != count) {
	d_cpPtr(&nRight, &ep1->e_down);
	pRight = d_bxGet(d_RelPtr, &nRight);
	if (pRight->x_count != 1) {
	    d_cpPtr(&nLeft, &node2);
	    pLeft = *pp;
	    COPY(&pLeft->x_entries[0].e_key[0], &ep1->e_key[0], d_KeyLen);
	    pLeft->x_entries[0].e_tuple.t_long = ep1->e_tuple.t_long;
	    COPY(&ep1->e_key[0], &pRight->x_entries[0].e_key[0], d_KeyLen);
	    ep1->e_tuple.t_long= pRight->x_entries[0].e_tuple.t_long;
	    d_cpPtr(&pLeft->x_entries[0].e_down, &pRight->x_left);
	    d_cpPtr(&pRight->x_left, &pRight->x_entries[0].e_down);
	    pLeft->x_count = 1;
	    d_bxMark(pLeft, &nLeft);
	    d_bxMark(n, node);
	    j = pRight->x_count - 1;
	    pRight->x_count = j;
	    ep1 = &pRight->x_entries[0];
	    while (j-- != 0) {
		ep2 = (ENTRY *)((byte *)ep1 + d_EntLen);
		COPY(ep1, ep2, d_EntLen);
		ep1 = ep2;
	    }
	    d_bxMark(pRight, &nRight);
	    return(false);
	}
    }
    if (!atLeft) {
	if (i == 1) {
	    d_cpPtr(&nLeft, &n->x_left);
	} else {
	    d_cpPtr(&nLeft, &((ENTRY *)((byte *)ep2 - d_EntLen))->e_down);
	}
	pLeft = d_bxGet(d_RelPtr, &nLeft);
	if (pLeft->x_count != 1) {
	    d_cpPtr(&nRight, &node2);
	    pRight = *pp;
	    j = pLeft->x_count - 1;
	    pLeft->x_count = j;
	    ep1 = (ENTRY *)((byte *)&pLeft->x_entries[0] + d_EntLen * j);
	    COPY(&pRight->x_entries[0].e_key[0],&ep2->e_key[0], d_KeyLen);
	    pRight->x_entries[0].e_tuple.t_long= ep2->e_tuple.t_long;
	    COPY(&ep2->e_key[0], &ep1->e_key[0], d_KeyLen);
	    ep2->e_tuple.t_long = ep1->e_tuple.t_long;
	    d_cpPtr(&pRight->x_entries[0].e_down, &pRight->x_left);
	    d_cpPtr(&pRight->x_left, &ep1->e_down);
	    pRight->x_count = 1;
	    d_bxMark(pRight, &nRight);
	    d_bxMark(n, node);
	    d_bxMark(pLeft, &nLeft);
	    return(false);
	}
    }

    /* no sibling node can contribute an entry. We must merge
       the empty node, an entry from the current (parent) node,
       and a sibling into one node: */

    if (i != count) {
	d_cpPtr(&nLeft, &node2);
	pLeft = *pp;
	d_cpPtr(&nRight, &ep1->e_down);
	pRight = d_bxGet(d_RelPtr, &nRight);
	COPY((ENTRY *)((byte *)&pRight->x_entries[0]+d_EntLen), &pRight->x_entries[0], d_EntLen);
	pRight->x_count = 2;
	COPY(&pRight->x_entries[0].e_key[0],&ep1->e_key[0], d_KeyLen);
	pRight->x_entries[0].e_tuple.t_long= ep1->e_tuple.t_long;
	d_cpPtr(&pRight->x_entries[0].e_down, &pRight->x_left);
	d_cpPtr(&pRight->x_left, &pLeft->x_left);
	if (atLeft) {
	    d_cpPtr(&n->x_left, &nRight);
	    i = 1;
	    ep2 = ep1;
	    ep1 = (ENTRY *)((byte *)ep1 + d_EntLen);
	} else {
	    COPY(&ep1->e_key[0], &ep2->e_key[0], d_KeyLen);
	    ep1->e_tuple.t_long = ep2->e_tuple.t_long;
	}
	if (d_Inode) {
	    d_rnFree(d_RelPtr, &nLeft);
	} else {
	    d_rxFree(d_RelPtr, &nLeft);
	}
	d_bxMark(pRight, &nRight);
	while (i++ != count) {
	    COPY(ep2, ep1, d_EntLen);
	    ep2 = ep1;
	    ep1 = (ENTRY *)((byte *)ep1 + d_EntLen);
	}
    } else { /* NOTE: i == count !!! */
	d_cpPtr(&nRight, &node2);
	pRight = *pp;
	if (i == 1) {
	    d_cpPtr(&nLeft, &n->x_left);
	} else {
	    d_cpPtr(&nLeft,&((ENTRY *)((byte *)ep2 - d_EntLen))->e_down);
	}
	pLeft = d_bxGet(d_RelPtr, &nLeft);
	pLeft->x_count = 2;
	ep1 = (ENTRY *)((byte *)&pLeft->x_entries[0] + d_EntLen);
	COPY(&ep1->e_key[0], &ep2->e_key[0], d_KeyLen);
	ep1->e_tuple.t_long = ep2->e_tuple.t_long;
	d_cpPtr(&ep1->e_down, &pRight->x_left);
	if (d_Inode) {
	    d_rnFree(d_RelPtr, &nRight);
	} else {
	    d_rxFree(d_RelPtr, &nRight);
	}
	d_bxMark(pLeft, &nLeft);
    }
    count--;
    n->x_count = count;
    if (count == 0) {
	/* as before, we refresh the returned node: */
	*pp = d_bxGet(d_RelPtr, node);
	return(true);
    } else {
	d_bxMark(n, node);
	return(false);
    }
}

/*
 * d_xDel0 - delete an entry from a B-tree. This routine must operate as the
 *	     reverse of the insertion routine, so that the assumptions about
 *	     the tree are not violated. In particular, empty nodes are not
 *	     allowed, thus an empty tree must free everything and leave the
 *	     root as a null pointer again. Return 'true' if the root node
 *	     has been modified. In that case, the pointer may be modified to
 *	     a null POINTER if the entire tree has vanished.
 */

d_xDel0(root)
    register POINTER *root;
{
    XBLOCK *n;
    POINTER temp;

    if (root->p_file == 0) {
	d_xGone();
	return(false);
    } else {
	if (d_xDel1(root, false, &n)) {

	    /* the root node has only one child; reduce height of tree: */

	    d_cpPtr(&temp, &n->x_left);
	    if (d_Inode) {
		d_rnFree(d_RelPtr, root);
	    } else {
		d_rxFree(d_RelPtr, root);
	    }
	    d_cpPtr(root, &temp);
	    return(true);
	} else {
	    return(false);
	}
    }
}
