#include "system.h"
#include "dbint.h"

/* procedures for use with relations. */

#define RCACHE	10

extern d_eInit();
extern d_print();
extern d_pNum();
extern d_error();
extern d_setErr();
extern d_abort();
extern byte *d_alloc();
extern d_free();

extern bool d_fopen();
extern d_fclose();
extern byte d_fget();
extern d_fread();

extern d_bInit();
extern d_bTerm();
extern byte *d_bOpen();
extern d_bCls1();
extern d_bCls2();
extern byte *d_bxGet();
extern d_bsNew();
extern d_bcNew();
extern d_bxMark();
extern d_bdPut();
extern d_bpStuff();
extern byte d_bdGet();
extern d_bgStuff();
extern byte d_bxhGet();
extern d_bxhPut();

extern d_scInit();
extern d_scTerm();
extern d_xsMod();
extern d_putTup();
extern d_uInd();

extern d_xInit();
extern d_xTerm();

byte d_InitF = 0;

RELATION *d_RCache[RCACHE];

DATABASE *d_DBList;
word d_DBCount;

RELATION *d_RelList;
word d_RelCnt;

bool d_Alloced;

int (*d_Logger)();

/*
 * dbInit - routine that must be called to initialize database operations.
 */

dbInit()
{
    d_print("ULTIMATUM routines (V");
    d_pNum(VERSION / 10);
    d_print(".");
    d_pNum(VERSION % 10);
    d_print(VARIANT);
    d_print("): Copyright (C) 1985 by Northware Systems Corporation\r\n");
    d_eInit();
    d_bInit();
    d_xInit();
    d_scInit();
    d_DBList = nil;
    d_RelList = nil;
    d_DBCount = 0;
    d_RelCnt = 0;
    d_Logger = nil;
    d_InitF = 0xff;
}

/*
 * setLog - set in a log routine.
 */

setLog(logger)
    int (*logger)();
{
    d_Logger = logger;
}

/*
 * d_pfLog - log field puts.
 */

d_pfLog(relNum, tuple, fldNum, typ, len, ptr)
    word relNum;
    long tuple;
    int fldNum;
    byte typ;
    word len;
    byte *ptr;
{
    if (d_Logger != nil) {
	(*d_Logger)(3, relNum, tuple, fldNum, typ, len, ptr);
    }
}

/*
 * d_dGet - get the internal structure of a dictionary from the .DIC file.
 *	    Return 'true' if all OK, else return 'false'.
 */

bool
d_dGet(name, d)
    char *name;
    DICT *d;
{
    FIELD *f, *f1;
    char nameBuff[15];
    word c, i;
    char *p;
    char ch;

    if (!d_fopen(0, name, "DIC", true)) {
	d_setErr(8);
	return(false);
    } else {
	if (d_fget() != VERSION) {
	    d_abort(7);
	}
	d_fread(&c, sizeof(word));
	d->d_fldCnt = c;
	f = (FIELD *)d_alloc(c * sizeof(FIELD));
	f1 = f;
	d->d_flds = f;
	while (c-- != 0) {
	    f->f_type = d_fget();
	    d_fread(&f->f_len, sizeof(word));
	    p = &nameBuff[0];
	    do {
		ch = d_fget();
		*p++ = ch;
	    } while (ch != 0);
	    p = (char *)d_alloc(strlen(&nameBuff[0]) + 1);
	    f++->f_name = p;
	    strcpy(p, &nameBuff[0]);
	}
	d_fclose();
	return(true);
    }
}

/*
 * dbOpen - user routine to open a database. The database (dictionary) name
 *	    is passed, and the databaseID (1 - n) is returned. 0 is returned
 *	    if something goes wrong.
 */

word
dbOpen(name)
    char *name;
{
    DATABASE *db, *fdb, *tdb, *prev;
    word dbNum, n;
    DICT d;

    if (d_InitF != 0xff) {
	d_abort(20);
    }
    fdb = nil;
    tdb = nil;
    prev = nil;
    db = d_DBList;
    n = 0;
    while (db != nil) {
	n++;
	if (strcmp(db->db_name, name) == 0 && db->db_act) {
	    tdb = db;
	} else if (!db->db_act) {
	    fdb = db;
	    dbNum = n;
	}
	prev = db;
	db = db->db_next;
    }
    if (tdb != nil) {
	d_error(9);
    }
    if (!d_dGet(name, &d)) {
	return(0);
    } else {
	if (fdb == nil) {
	    db = (DATABASE *)d_alloc(sizeof(DATABASE));
	    db->db_next = nil;
	    if (prev == nil) {
		d_DBList = db;
	    } else {
		prev->db_next = db;
	    }
	    dbNum = ++d_DBCount;
	} else {
	    db = fdb;
	}
	db->db_name = name;
	db->db_dict.d_flds = d.d_flds;
	db->db_dict.d_fldCnt = d.d_fldCnt;
	db->db_act = true;
	return(dbNum);
    }
}

/*
 * d_dpget2 - get a pointer to the appropriate DATABASE.
 */

DATABASE *
d_dpget2(dbNum)
    word dbNum;
{
    DATABASE *db;

    if (dbNum == 0 || dbNum > d_DBCount) {
	d_error(10);
    }
    db = d_DBList;
    while (--dbNum != 0) {
	db = db->db_next;
    }
    if (!db->db_act) {
	d_error(11);
    }
    return(db);
}

/*
 * d_rxgword - local routine to get a word from the index header.
 */

word
d_rxgword()
{
    word low;

    low = d_bxhGet();
    return(low + (d_bxhGet() << 8));
}

/*
 * d_rxgptr - local routine to get a POINTER from the index header.
 */

d_rxgptr(p)
    POINTER *p;
{
    p->p_block = d_rxgword();
    p->p_file = d_bxhGet();
    p->p_pad = 0xff;
}

/*
 * d_rxglong - local routine to get a long from the index header.
 */

long
d_rxglong()
{
    long low;

    low = d_rxgword();
    return(low + ((long)d_rxgword() << 16));
}

/*
 * rOpen - user routine to open a relation. The relation name is passed,
 *	   along with the database number. A relationID is returned. 0 is
 *	   returned if the relation cannot be opened.
 */

word
rOpen(dbNum, drive, name)
    word dbNum;
    char drive;
    char *name;
{
    RELATION *r, *fr, *tr, *prev;
    DATABASE *db;
    INDEX *x;
    INDEX **xNext;
    RFIELD *flds;
    byte *rio;
    long pos;
    word relNum, n;
    short f;

    if (d_InitF != 0xff) {
	d_abort(20);
    }
    db = d_dpget2(dbNum);
    fr = nil;
    tr = nil;
    prev = nil;
    r = d_RelList;
    n = 0;
    while (r != nil) {
	n++;
	if (r->r_dbase == dbNum && strcmp(r->r_name,name) == 0 && r->r_act) {
	    tr = r;
	} else if (!r->r_act) {
	    fr = r;
	    relNum = n;
	}
	prev = r;
	r = r->r_next;
    }
    if (fr == nil) {
	relNum = d_RelCnt + 1;
    }
    if (tr != nil) {
	d_error(12);
    }
    if (fr == nil) {
	r = (RELATION *)d_alloc(sizeof(RELATION));
	r->r_next = nil;
	if (prev == nil) {
	    d_RelList = r;
	} else {
	    prev->r_next = r;
	}
	d_RelCnt++;
	if (relNum < RCACHE) {
	    d_RCache[relNum - 1] = r;
	}
    } else {
	r = fr;
    }
    r->r_name = (char *)d_alloc(strlen(name) + 1);
    strcpy(r->r_name, name);
    r->r_dbase = dbNum;
    rio = d_bOpen(drive, r);
    if (rio == nil) {
	d_free(r->r_name, strlen(name) + 1);
	r->r_act = false;
	return(0);
    } else {
	r->r_riol = rio;
	r->r_act = true;
	r->r_change = false;
	r->r_dbptr = db;
	r->r_fSize = d_rxgword();
	r->r_tcurr = 0l;
	n = d_rxgword();
	r->r_fldCnt = n;
	flds = (RFIELD *)d_alloc(sizeof(RFIELD) * n);
	r->r_fields = flds;
	pos = 1l;		/* first byte is 'deleted' flag */
	while (n-- != 0) {
	    f = d_rxgword();
	    if (f >= db->db_dict.d_fldCnt) {
		d_abort(2);
	    }
	    flds->rf_pos = pos;
	    flds->rf_field = f;
	    flds->rf_scnCnt = 0;
	    flds->rf_index = false;
	    flds->rf_mod = false;
	    flds++;
	    pos += db->db_dict.d_flds[f].f_len;
	}
	if (pos < 4l) {
	    pos = 4l;
	}
	r->r_len = pos;
	r->r_tnext = d_rxglong();
	r->r_tfree = d_rxglong();
	d_rxgptr(&r->r_xnext);
	d_rxgptr(&r->r_xfree);
	d_rxgptr(&r->r_nfree);
	xNext = &r->r_index;
	n = d_rxgword();
	while (n-- != 0) {
	    x = (INDEX *)d_alloc(sizeof(INDEX));
	    *xNext = x;
	    xNext = &x->x_next;
	    x->x_unique = d_bxhGet();
	    x->x_type = d_bxhGet();
	    x->x_len = d_rxgword();
	    f = d_rxgword();
	    x->x_field = f;
	    d_rxgptr(&x->x_root);
	    if (f >= 0) {
		r->r_fields[f].rf_index = true;
	    }
	}
	*xNext = nil;
	return(relNum);
    }
}

/*
 * d_rCre - set up a RELATION for a relation being created.
 */

word
d_rCre(rp)
    RELATION **rp;
{
    RELATION *r, *prev;
    word relNum;

    if (d_InitF != 0xff) {
	d_abort(20);
    }
    r = d_RelList;
    relNum = 1;
    prev = nil;
    while (r != nil && r->r_act) {
	relNum++;
	prev = r;
	r = r->r_next;
    }
    if (r == nil) {
	r = (RELATION *)d_alloc(sizeof(RELATION));
	r->r_next = nil;
	if (prev == nil) {
	    d_RelList = r;
	} else {
	    prev->r_next = r;
	}
	d_RelCnt++;
	if (relNum < RCACHE) {
	    d_RCache[relNum - 1] = r;
	}
    }
    *rp = r;
    return(relNum);
}

/*
 * d_rxpword - local routine to put a word to the index header.
 */

d_rxpword(w)
    word w;
{
    d_bxhPut(w);
    d_bxhPut(w >> 8);
}

/*
 * d_rxpptr - local routine to put a POINTER to the index header.
 */

d_rxpptr(p)
    POINTER *p;
{
    d_rxpword(p->p_block);
    d_bxhPut(p->p_file);
}

/*
 * d_rxplong - local routine to put a long to the index header.
 */

d_rxplong(n)
    long n;
{
    d_rxpword(n);
    d_rxpword(n >> 16);
}

/*
 * d_rpget2 - get a RELATION pointer from a relationID.
 */

RELATION *
d_rpget2(relNum)
    word relNum;
{
    RELATION *r;

    if (relNum == 0 || relNum > d_RelCnt) {
	d_error(13);
    }
    if (relNum < RCACHE) {
	r = d_RCache[relNum - 1];
    } else {
	r = d_RelList;
	while (--relNum != 0) {
	    r = r->r_next;
	}
    }
    if (!r->r_act) {
	d_error(14);
    }
    return(r);
}

/*
 * d_rCls2 - half of the rClose code.
 */

d_rCls2(r)
    RELATION *r;
{
    INDEX *x;
    RFIELD *flds;
    word n;

    d_rxpword(r->r_fSize);
    flds = r->r_fields;
    n = r->r_fldCnt;
    d_rxpword(n);
    while (n-- != 0) {
	d_rxpword(flds++->rf_field);
    }
    d_rxplong(r->r_tnext);
    d_rxplong(r->r_tfree);
    d_rxpptr(&r->r_xnext);
    d_rxpptr(&r->r_xfree);
    d_rxpptr(&r->r_nfree);
    n = 0;
    x = r->r_index;
    while (x != nil) {
	x = x->x_next;
	n++;
    }
    d_rxpword(n);
    x = r->r_index;
    while (x != nil) {
	d_bxhPut(x->x_unique);
	d_bxhPut(x->x_type);
	d_rxpword(x->x_len);
	d_rxpword(x->x_field);
	d_rxpptr(&x->x_root);
	x = x->x_next;
    }
}

/*
 * rClose - user routine to close a relation.
 */

rClose(relNum)
    word relNum;
{
    RELATION *r;
    INDEX *x, *xt;

    r = d_rpget2(relNum);
    d_putTup(r);
    if (r->r_change) {
	d_Alloced = false;
	d_bCls1(r);
	d_rCls2(r);
	if (d_Alloced) {
	    d_bCls1(r);
	    d_rCls2(r);
	}
    }
    d_bCls2(r);
    r->r_act = false;
    /* this comment stops the CPM compiler from dieing!! */
    d_free((byte *)r->r_name, strlen(r->r_name) + 1);
    d_free((byte *)r->r_fields, sizeof(RFIELD) * r->r_fldCnt);
    x = r->r_index;
    while (x != nil) {
	xt = x;
	x = x->x_next;
	d_free(xt, sizeof(INDEX));
    }
}

/*
 * dbClose - user callable routine to close an entire database. Mostly just
 *	     marks the entry as non-active and closes all the relations.
 */

dbClose(dbNum)
    word dbNum;
{
    DATABASE *db;
    RELATION *r;
    FIELD *f;
    word relNum;

    db = d_dpget2(dbNum);
    relNum = 0;
    r = d_RelList;
    while (r != nil) {
	relNum++;
	if (r->r_act && r->r_dbase == dbNum) {
	    if (!rClose(relNum)) {
		d_abort(5);
	    }
	}
	r = r->r_next;
    }
    f = db->db_dict.d_flds;
    for (relNum = 0; relNum != db->db_dict.d_fldCnt; relNum++) {
	d_free(f->f_name, strlen(f->f_name) + 1);
	f++;
    }
    d_free(db->db_dict.d_flds, sizeof(FIELD) * db->db_dict.d_fldCnt);
    db->db_act = false;
}

/*
 * dbTerm - routine called to terminate database activities and free space.
 */

dbTerm()
{
    DATABASE *db, *dbt;
    RELATION *r, *rt;
    word dbNum;

    dbNum = 0;
    db = d_DBList;
    while (db != nil) {
	dbNum++;
	if (db->db_act) {
	    dbClose(dbNum);
	}
	db = db->db_next;
    }
#ifdef CLEANUP
    db = d_DBList;
    while (db != nil) {
	dbt = db;
	db = db->db_next;
	d_free(dbt, sizeof(DATABASE));
    }
    r = d_RelList;
    while (r != nil) {
	rt = r;
	r = r->r_next;
	d_free(rt, sizeof(RELATION));
    }
    d_scTerm();
    d_xTerm();
    d_bTerm();
#endif
}

/*
 * d_rxAlloc - allocate a new index block in a relation.
 */

d_rxAlloc(r, p)
    RELATION *r;
    POINTER *p;
{
    byte *b;

    d_Alloced = true;
    r->r_change = true;
    if (r->r_xfree.p_file != 0) {
	d_cpPtr(p, &r->r_xfree);
	b = d_bxGet(r, p);
	d_cpPtr(&r->r_xfree, b);
    } else {
	d_cpPtr(p, &r->r_xnext);
	r->r_xnext.p_block += BLKSIZE / 128;
	if (r->r_xnext.p_block == r->r_fSize * (1024 / 128)) {
	    r->r_xnext.p_block = 0;
	    r->r_xnext.p_file++;
	    if (r->r_xnext.p_file <= 0) {
		d_abort(6);
	    }
	}
	/* force writing of new block if not already present in file: */
	d_bsNew();
	b = d_bxGet(r, p);
	d_bcNew();
    }
}

/*
 * d_rxFree - free an index block in the given relation.
 */

d_rxFree(r, p)
    RELATION *r;
    POINTER *p;
{
    byte *b;

    r->r_change = true;
    b = d_bxGet(r, p);
    d_cpPtr(b, &r->r_xfree);
    d_bxMark(b, p);
    d_cpPtr(&r->r_xfree, p);
}

/*
 * d_rnAlloc - allocate a new small inode block in the given relation.
 */

d_rnAlloc(r, p)
    RELATION *r;
    POINTER *p;
{
    byte *b, *b1;
    POINTER ptr;
    byte cnt;

    r->r_change = true;
    if (r->r_nfree.p_file != 0) {
	d_cpPtr(p, &r->r_nfree);
	ptr.p_file = p->p_file;
	ptr.p_block = p->p_block & (-(BLKSIZE / INOSIZE));
	b = d_bxGet(r, &ptr) +
	    (p->p_block & (BLKSIZE / INOSIZE - 1)) * INOSIZE;
	d_cpPtr(&r->r_nfree, b);
    } else {
	d_rxAlloc(r, p);
	b = d_bxGet(r, p);
	b1 = b + INOSIZE;
	d_cpPtr(&ptr, p);
	ptr.p_block++;
	d_cpPtr(&r->r_nfree, &ptr);
	for (cnt = BLKSIZE / INOSIZE - 2; cnt != 0; cnt--) {
	    ptr.p_block++;
	    d_cpPtr(b1, &ptr);
	    b1 += INOSIZE;
	}
	((POINTER *)b1)->p_file = 0;
	d_bxMark(b, p);
    }
}

/*
 * d_rnFree - free a small inode block.
 */

d_rnFree(r, p)
    RELATION *r;
    POINTER *p;
{
    byte *b;
    POINTER ptr;

    r->r_change = true;
    ptr.p_file = p->p_file;
    ptr.p_block = p->p_block & (-(BLKSIZE / INOSIZE));
    b = d_bxGet(r, &ptr) +
	(p->p_block & (BLKSIZE / INOSIZE - 1)) * INOSIZE;
    d_cpPtr(b, &r->r_nfree);
    d_bxMark(b, p);
    d_cpPtr(&r->r_nfree, p);
}

/*
 * tAlloc - allocate a new tuple in the given relation.
 */

long
tAlloc(relNum)
    word relNum;
{
    RELATION *r;
    RFIELD *rf;
    FIELD *f, *fp;
    long t, p1, p2;
    word cnt, len, i;

    r = d_rpget2(relNum);
    r->r_change = true;
    /* MUST call d_putTup before setting d_bsNew, since the call to
       d_putTup may create a new index block, which will result in
       d_bsNew being cleared too soon. */
    d_putTup(r);
    /* set all mod flags to indicate that the tuple has had all relevant
       (none in this case) un-indexing done. */
    d_xsMod(r, true);
    if (r->r_tfree > 0l) {
	t = r->r_tfree;
	d_bgStuff(r, (t - 1l) * r->r_len + 1l, &r->r_tfree, sizeof(long));
    } else {
	t = r->r_tnext;
	r->r_tnext = t + 1l;
	d_bsNew();
    }
    r->r_tcurr = t;
    f = r->r_dbptr->db_dict.d_flds;
    p1 = (t - 1l) * r->r_len;
    d_bdPut(r, p1, false);
    rf = r->r_fields;
    for (cnt = 0; cnt != r->r_fldCnt; cnt++) {
	p2 = p1 + rf->rf_pos;
	fp = &f[rf->rf_field];
	switch (fp->f_type) {
	case FTBOOL:
	    d_bdPut(r, p2, false);
	    break;
	case FTSHORT:
	    d_bpStuff(r, p2, "\0\0\0\0\0\0\0\0", sizeof(short));
	    break;
	case FTLONG:
	    d_bpStuff(r, p2, "\0\0\0\0\0\0\0\0", sizeof(long));
	    break;
	case FTFLOAT:
	    d_bpStuff(r, p2, "\0\0\0\0\0\0\0\0", sizeof(double));
	    break;
	case FTCHAR:
	    len = fp->f_len;
	    for (i = 0; i != len; i++) {
		d_bdPut(r, p2 + i, 0);
	    }
	    break;
	case FTSTRING:
	case FTUCSTR:
	    len = fp->f_len;
	    for (i = 0;
		    i != ((len + (BLKSIZE - 1)) / BLKSIZE) * BLKSIZE;
		    i += BLKSIZE) {
		d_bdPut(r, p2 + i, 0);
	    }
	    d_bdPut(r, p2 + (len - 1), 0);
	    break;
	default:
	    d_abort(8);
	}
	rf++;
    }
    d_bcNew();
    if (d_Logger != nil) {
	(*d_Logger)(1, relNum, t, 0, 0, 0, nil);
    }
    return(t);
}

/*
 * tFree - free a tuple in the given relation.
 */

tFree(relNum, tup)
    word relNum;
    long tup;
{
    RELATION *r;
    long pos;

    r = d_rpget2(relNum);
    if (tup < 1l || tup >= r->r_tnext) {
	d_error(15);
    }
    pos = (tup - 1l) * r->r_len;
    if (d_bdGet(r, pos)) {
	d_error(160);
    }
    r->r_tcurr = tup;
    d_uInd(r);
    r->r_change = true;
    d_bdPut(r, pos, true);
    d_bpStuff(r, pos + 1l, &r->r_tfree, sizeof(long));
    r->r_tfree = tup;
    r->r_tcurr = 0l;
    if (d_Logger != nil) {
	(*d_Logger)(2, relNum, tup, 0, 0, 0, nil);
    }
}

/*
 * chkTuple - check a tuple for validity.
 */

int
chkTuple(relNum, tup)
    word relNum;
    long tup;
{
    RELATION *r;
    long pos;

    r = d_rpget2(relNum);
    if (tup < 1l || tup >= r->r_tnext) {
	return(2);
    } else {
	pos = (tup - 1l) * r->r_len;
	return((d_bdGet(r, pos)) ? 1 : 0);
    }
}

/*
 * tNext - return the tupleid of the next unallocated tuple.
 */

long
tNext(relNum)
    word relNum;
{

    return(d_rpget2(relNum)->r_tnext);
}
