#include "system.h"
#include "dbint.h"

extern d_setErr();
extern d_error();
extern byte *d_alloc();
extern d_free();

extern DATABASE *d_dpget2();
extern word d_rCre();
extern byte *d_rbCre();

/* a structure to use for a list of fields: */

typedef struct flist {
    struct flist *fl_next;
    long fl_pos;
    int fl_field;
} FLIST;

FLIST *d_RCFList;
FLIST **d_RCFNxt;
int d_RCFCnt;
long d_RCFPos;
DATABASE *d_RCFDB;
RELATION *d_RCFRel;

/* routines for creating new relations */

word
rCreate(dbNum, name, xdrives, ddrives, fSize)
    word dbNum;
    char *name, *xdrives, *ddrives;
    word fSize;
{
    DATABASE *db;
    RELATION *r;
    word relNum;

    db = d_dpget2(dbNum);
    if ((relNum = strlen(xdrives)) <= 0 || relNum > 127 ||
	    (relNum = strlen(ddrives)) <= 0 || relNum > 127) {
    	d_setErr(17);
	return(0);
    } else if (fSize <= 0 || fSize > 8192) {
	d_setErr(33);
	return(0);
    } else {
	relNum = d_rCre(&r);
	r->r_name = (char *)d_alloc(strlen(name) + 1);
	strcpy(r->r_name, name);
	r->r_index = nil;
	r->r_dbptr = db;
	r->r_fSize = fSize;
	r->r_tnext = 1l;
	r->r_tfree = 0l;
	r->r_tcurr = 0l;
	r->r_dbase = dbNum;
	r->r_xnext.p_block = BLKSIZE / 128;
	r->r_xnext.p_file = 1;
	r->r_xnext.p_pad = 0xff;
	r->r_xfree.p_file = 0;
	r->r_nfree.p_file = 0;
	r->r_act = true;
	r->r_change = true;		/* force write of header */
	r->r_riol = d_rbCre(r, xdrives, ddrives);
	d_RCFCnt = 0;
	d_RCFPos = 1l;
	d_RCFDB = db;
	d_RCFRel = r;
	d_RCFList = nil;
	d_RCFNxt = &d_RCFList;
	if (r->r_riol == nil) {
	    r->r_act = false;
	    d_free((byte *)r->r_name, strlen(name) + 1);
	    return(0);
	} else {
	    return(relNum);
	}
    }
}

/*
 * rcField - add a field to the relation being created.
 */

bool
rcField(fNum)
    int fNum;
{
    FLIST *fl;

    if (!d_RCFRel->r_act) {
	return(false);
    } else if (fNum <= 0 || fNum > d_RCFDB->db_dict.d_fldCnt) {
	d_error(20);
    } else {
	fNum--;
	fl = d_RCFList;
	while (fl != nil && fl->fl_field != fNum) {
	    fl = fl->fl_next;
	}
	if (fl != nil) {
	    d_setErr(21);
	    return(false);
#ifdef PARANOID
	} else if (d_RCFCnt == 0x7fff) {
	    d_setErr(38);
	    return(false);
	} else if (0x7fffffffl - d_RCFPos <
		   d_RCFDB->db_dict.d_flds[fNum].f_len) {
	    d_setErr(39);
	    return(false);
#endif
	} else {
	    d_RCFCnt++;
	    fl = (FLIST *)d_alloc(sizeof(FLIST));
	    fl->fl_next = nil;
	    fl->fl_pos = d_RCFPos;
	    fl->fl_field = fNum;
	    *d_RCFNxt = fl;
	    d_RCFNxt = &fl->fl_next;
	    d_RCFPos += d_RCFDB->db_dict.d_flds[fNum].f_len;
	    return(true);
	}
    }
}

/*
 * rcEnd - end creation of a new relation.
 */

bool
rcEnd()
{
    RFIELD *flds;
    FLIST *t;

    if (!d_RCFRel->r_act) {
	return(false);
    } else if (d_RCFCnt == 0) {
	d_setErr(22);
	return(false);
    } else {
	flds = (RFIELD *)d_alloc(d_RCFCnt * sizeof(RFIELD));
	d_RCFRel->r_fields = flds;
	d_RCFRel->r_fldCnt = d_RCFCnt;
	while (d_RCFList != nil) {
	    flds->rf_pos = d_RCFList->fl_pos;
	    flds->rf_field = d_RCFList->fl_field;
	    flds->rf_scnCnt = 0;
	    flds->rf_index = false;
	    flds->rf_mod = false;
	    flds++;
	    t = d_RCFList;
	    d_RCFList = d_RCFList->fl_next;
	    d_free((byte *)t, sizeof(FLIST));
	}
	if (d_RCFPos < 4l) {
	    d_RCFPos = 4l;
	}
	d_RCFRel->r_len = d_RCFPos;
	return(true);
    }
}
