#include "system.h"
#include "db.h"

#define FILESIZE	64	/* size in K bytes of each data/index file */
#undef  DICCRE			/* create TEST.DIC initially, else use it */
#define FIELD8			/* virtual integer field */
#define FIELD9			/* virtual string[30] field */
#define CREATE			/* create the relation first */
#define ADD			/* add tuples first (the .'s) */
#undef  REOPEN			/* close after add, then re-open */
#define DELETE			/* delete tuples when done (the -'s) */
#undef  LOGGING			/* log field updates */
#define ISCAN			/* do index scanning */
#define SETALL			/* go down into inodes on scan */
#define BRIEFSCAN		/* use short scan (# only, using fNext) */
#undef  INTERACTIVE		/* use interactive scan (n, p, q) */
#undef  STARTSCAN		/* use interactive scan start positions */
#undef  CHECK			/* verify existence of tuples (the ?'s) */
#define UNIQUE	false		/* true or false flag to xCreate */
#define MOD10			/* apply mod 10 to key values */
#define RERUN			/* run everything twice */
#define COUNT	175		/* number of tuples to play with */
#undef  STATS			/* collect stats (must be in DBIO.C also) */

/* Some combinations of the above don't make much sense.
   For example, MOD10 with UNIQUE = true will result in very few of the
   tuples being indexed (earlier occurrences are thrown away).

   If SETALL is defined, then BRIEFSCAN must be also, since if we go down
   into the inodes, we can't return keys, hence scNext/scPrev won't run. */

#define BUFFLEN	40
char Buffer[BUFFLEN + 1 + 2];
long GTuple, GLong;
int GInt;
double GFloat;
bool GBool;
char *BuffPtr;

word Num, Seed, FldLen;

#ifdef STATS
pLong(mess, n)
    char *mess;
    long n;
{
    char *p;

    d_print(mess);
    p = &Buffer[10];
    *p = '\0';
    do {
	*--p = n % 10l + '0';
	n /= 10l;
    } while (n != 0l);
    d_print(p);
    d_print("\r\n");
}
#endif

hex(buff)
    char *buff;
{
    word n;
    int i;
    char digit;

    n = Num;
    buff += 4;
    for (i = 4; i != 0; i--) {
	digit = n & 0x0f;
	if (digit > 9) {
	    digit += 'a' - 10;
	} else {
	    digit += '0';
	}
	*--buff = digit;
	n >>= 4;
    }
}

pTupId(tuple)
long tuple;
{
    long z;
    
    z = 0L;
    if (tuple < z) {
	d_print("**");
    } else {
	d_pNum((word)tuple);
    }
}

#ifdef CHECK
#ifdef MOD10

checka(flag, i, f)
    bool flag;
    int i, f;
{
    long tuple, z;

    d_print("Scan of ");
    d_pNum(i);
    d_print(", field ");
    d_pNum(f);
    if (flag) {
	d_print(":\r\n");
	z = 0L;
	while ((tuple = fNext(1)) != z) {
	    pTupId(tuple);
	    d_print(" ");
	}
	d_print("\r\n");
    } else {
	d_print(" - none found\r\n");
    }
}

#else

check(tuple, i, f)
    long tuple;
    int i, f;
{
    if ((int)tuple != i) {
	d_print("Lookup failed, tuple = ");
	d_pNum((int) tuple);
	d_print(", field ");
	d_pNum(f);
	d_print("\r\n");
    }
}

#endif

#endif

#ifdef ISCAN

#ifdef INTERACTIVE

#ifdef STARTSCAN

sChk(any)
bool any;
{
    if (!any) {
	d_print("scanXXX failed\r\n");
    }
}

#ifdef BRIEFSCAN
long
getPos(fld)
    int fld;
#else
word
getPos(fld, t)
    int fld;
    long *t;
#endif
{
    word n;
    char *p;
    bool doPrev;

    d_print("> ");
#ifdef CPM
    Buffer[0] = FldLen + 1;
    bdos(10, Buffer);
    d_print("\r\n");
    Buffer[Buffer[1] + 2] = '\0';
#else
    blah - read characters into Buffer[2 -> BUFFLEN+1]
#endif
    if (Buffer[2] == '\0' || Buffer[2] == 'q' && Buffer[3] == '\0') {
	/* empty line - stop scanning this field */
	fAbort(1);
#ifdef BRIEFSCAN
	return(0l);
#else
	return(0);
#endif
    }
    if (Buffer[2] == 'n' && Buffer[3] == '\0') {
	/* just go to next in scan */
#ifdef BRIEFSCAN
	return(fNext(1));
#else
	return(scNext(1, BuffPtr, t));
#endif
    }
    if (Buffer[2] == 'p' && Buffer[3] == '\0') {
	/* just go to previous in scan */
#ifdef BRIEFSCAN
	return(fPrev(1));
#else
	return(scPrev(1, BuffPtr, t));
#endif
    }
    doPrev = false;
    p = &Buffer[2];
    if (*p == 'n') {
	++p;
    } else if (*p == 'p') {
	++p;
	doPrev = true;
    }
    if (fld == 1 || fld == 5 || fld == 7) {
	n = 0;
	while (*p >= '0' && *p <= '9') {
	    n = n * 10 + (*p++ - '0');
	}
    }
    fAbort(1);
    switch (fld) {
    case 1:
	sChk(scanInt(1, 1, 1, n));
	break;
    case 2:
    case 3:
	sChk(scanString(1, 1, fld, p));
	break;
    case 4:
	sChk(scanChars(1, 1, 4, p));
	break;
    case 5:
	sChk(scanFloat(1, 1, 5, (double)n));
	break;
    case 6:
	sChk(scanBool(1, 1, 6, (*p == 't')));
	break;
    case 7:
	sChk(scanLong(1, 1, 7, (long)n));
	break;
    default:
	d_print("Bad case");
	exit(1);
    }
#ifdef SETALL
    setAll(1);
#endif
#ifdef BRIEFSCAN
    return(doPrev ? fPrev(1) : fNext(1));
#else
    return(doPrev ? scPrev(1, BuffPtr, t) : scNext(1, BuffPtr, t));
#endif
}

/* matches #ifdef STARTSCAN: */
#else

#ifdef BRIEFSCAN

long
intN()
{
    long tuple;
    char ch;

    while (true) {
#ifdef CPM
	ch = bdos(1, 0);
#else
	ch = ???
#endif
	if (ch == 'n') {
	    return(fNext(1));
	}
	if (ch == 'p') {
	    return(fPrev(1));
	}
	if (ch == 'q') {
	    fAbort(1);
	    return(0l);
	}
	d_print("n, p, or q\r\n");
    }
}

/* matches #ifdef BRIEFSCAN: */
#else

word
intN(t)
    long *t;
{
    word depth;
    char ch;

    while (true) {
#ifdef CPM
	ch = bdos(1, 0);
#else
	ch = ???
#endif
	if (ch == 'n') {
	    return(scNext(1, BuffPtr, t));
	}
	if (ch == 'p') {
	    return(scPrev(1, BuffPtr, t));
	}
	if (ch == 'q') {
	    fAbort(1);
	    return(0);
	}
	d_print("n, p, or q\r\n");
    }
}

/* end of #ifdef BRIEFSCAN: */
#endif

/* end of #ifdef STARTSCAN: */
#endif

/* end of #ifdef INTERACTIVE: */
#endif

checki1(fld, typ, dir, mess)
    int fld;
    word typ;
    bool dir;
    char *mess;
{
    long t, l;
    word depth;

    d_print(mess);
    d_print(" index scan for field ");
    d_pNum(fld);
#ifdef STARTSCAN
    if (strtLow(1, 1, fld)) {
#else
    if (dir ? strtLow(1, 1, fld) : strtHigh(1, 1, fld)) {
#endif
#ifdef SETALL
	setAll(1);
#endif
	d_print(":\r\n");
#ifdef BRIEFSCAN
	l = 0L;
#ifdef INTERACTIVE
#ifdef STARTSCAN
	while ((t = getPos(fld)) != l) {
#else
	while ((t = intN()) != l) {
#endif
#else
	while ((t = (dir ? fNext(1) : fPrev(1))) != l) {
#endif
	    pTupId(t);
	    d_print(" ");
	}
	d_print("\r\n");
/* matches with #ifdef BRIEFSCAN: */
#else
	switch (typ) {
	case FTBOOL:
	    BuffPtr = (char *)&GBool;
	    break;
	case FTINT:
	    BuffPtr = (char *)&GInt;
	    FldLen = 5;
	    break;
	case FTLONG:
	    BuffPtr = (char *)&GLong;
	    FldLen = 5;
	    break;
	case FTFLOAT:
	    BuffPtr = (char *)&GFloat;
	    FldLen = 5;
	    break;
	case FTCHAR:
	    BuffPtr = &Buffer[0];
	    break;
	case FTSTRING:
	case FTUCSTR:
	    --FldLen;
	    Buffer[FldLen] = '\0';
	    BuffPtr = &Buffer[0];
	    break;
	}
#ifdef INTERACTIVE
#ifdef STARTSCAN
	while ((depth = getPos(fld, &t)) != 0) {
#else
	while ((depth = intN(&t)) != 0) {
#endif
#else
	while ((depth = (dir ?
			 scNext(1, BuffPtr, &t) :
			 scPrev(1, BuffPtr, &t)
			)) != 0) {
#endif
	    while (--depth != 0) {
		d_print(" - ");
	    }
	    pTupId(t);
	    d_print(": ");
	    switch (typ) {
	    case FTBOOL:
		d_print(GBool ? "t" : "f");
		break;
	    case FTINT:
		d_pNum(GInt);
		break;
	    case FTLONG:
		d_pNum((word)GLong);
		break;
	    case FTFLOAT:
		d_pNum((word)GFloat);
		d_print(".");
		break;
	    case FTCHAR:
	    case FTSTRING:
	    case FTUCSTR:
		Buffer[FldLen] = '\0';
		d_print(&Buffer[0]);
		break;
	    }
	    d_print("\r\n");
	}
/* end of #ifdef BRIEFSCAN: */
#endif
    } else {
	d_print(" - none\r\n");
    }
}

checki(fld, typ, len)
    int fld;
    word typ;
    word len;
{

    FldLen = len;
#ifdef STARTSCAN
    checki1(fld, typ, false, "");
#else
    checki1(fld, typ, true, "Forward");
    checki1(fld, typ, false, "Reverse");
#endif
}

/* end of #ifdef ISCAN: */
#endif

f2()
{
    strcpy(&Buffer[0], "XXXX is a name for a customer");
    hex(&Buffer[0]);
}

f3()
{
    strcpy(&Buffer[0], "XXXX is an address for a customer      ");
    hex(&Buffer[0]);
}

f4()
{
    strcpy(&Buffer[0], "XXXXxyz");
    hex(&Buffer[0]);
}

#ifdef FIELD9
f9()
{
    strcpy(&Buffer[0], "XXXX is a code for a customer");
    hex(&Buffer[0]);
}
#endif

rStep()
{
    Seed = Seed * 2113 + 13367;
    Seed = (Seed >> 8) + (Seed << 8);
}

run()
{
    long tuple;
    word i, j;

#ifdef ADD
    Seed = 4369;
    for (i = 1; i != COUNT + 1; i++) {
#ifdef MOD10
	Num = Seed % 10;
#else
	Num = Seed;
#endif
	tuple = tAlloc(1);
	d_pNum((word)tuple);
	pfInt(1, tuple, 1, Num);
	f2();
	pfString(1, tuple, 2, &Buffer[0]);
	f3();
	pfString(1, tuple, 3, &Buffer[0]);
	f4();
	pfChar(1, tuple, 4, &Buffer[0]);
	pfFloat(1, tuple, 5, (double)Num);
	pfBool(1, tuple, 6, true);
	pfLong(1, tuple, 7, (long)Num);
#ifdef FIELD8
	insInt(1, tuple, -1, Num * 2);
#endif
#ifdef FIELD9
	f9();
	insString(1, tuple, -2, &Buffer[0]);
#endif
	rUpdate(1);
	d_print(".");
	rStep();
    }
    d_print("\r\n");
#endif
#ifdef REOPEN
    rClose(1);
    rOpen(1, '0', "CUST");
#endif

#ifdef ISCAN
	checki(1, FTINT, 2);
	checki(2, FTUCSTR, 30);
	checki(3, FTSTRING, 40);
	checki(4, FTCHAR, 7);
	checki(5, FTFLOAT, 8);
	checki(6, FTBOOL, 1);
	checki(7, FTLONG, 4);
#ifdef FIELD8
	checki(-1, FTINT, 2);
#endif
#ifdef FIELD9
	checki(-2, FTUCSTR, 30);
#endif
#endif
    Seed = 4369;
#ifdef CHECK
#ifdef MOD10
	for (Num = 0; Num != 10; Num++) {
	    checka(fAllInt(1, 1, 1, Num), Num, 1);
	    f2();
	    checka(fAllString(1, 1, 2, &Buffer[0]), Num, 2);
	    f3();
	    checka(fAllString(1, 1, 3, &Buffer[0]), Num, 3);
	    f4();
	    checka(fAllChar(1, 1, 4, &Buffer[0]), Num, 4);
	    checka(fAllFloat(1, 1, 5, (double)Num), Num, 5);
	    checka(fAllLong(1, 1, 7, (long)Num), Num, 7);
#ifdef FIELD8
	    checka(fAllInt(1, 1, -1, Num * 2), Num, -1);
#endif
#ifdef FIELD9
	    f9();
	    checka(fAllString(1, 1, -2, &Buffer[0]), Num, -2);
#endif
	}
	checka(fAllBool(1, 1, 6, true), 99, 6);
	checka(fAllBool(1, 1, 6, false), 99, 6);

/* matches with #ifdef MOD10: */
#else
	for (i = 1; i != COUNT + 1; i++) {
	    Num = Seed;
	    d_pNum(i);
	    check(fOneInt(1, 1, Num), i, 1);
	    f2();
	    check(fOneString(1, 2, &Buffer[0]), i, 2);
	    f3();
	    check(fOneString(1, 3, &Buffer[0]), i, 3);
	    f4();
	    check(fOneChar(1, 4, &Buffer[0]), i, 4);
	    check(fOneFloat(1, 5, (double)Num), i, 5);
	    check(fOneLong(1, 7, (long)Num), i, 7);
#ifdef FIELD8
	    check(fOneInt(1, -1, Num * 2), i, -1);
#endif
#ifdef FIELD9
	    f9();
	    check(fOneString(1, -2, &Buffer[0]), i, -2);
#endif
	    d_print("?");
	    rStep();
	}
	d_print("\r\n");
/* end of #ifdef MOD10: */
#endif

/* end of #ifdef CHECK: */
#endif

#ifdef DELETE

	for (i = COUNT; i != 0; i--) {
#ifdef MOD10
	    Num = Seed % 10;
#else
	    Num = Seed;
#endif
	    tuple = i;
	    d_pNum(i);
	    tFree(1, tuple);
#ifdef FIELD8
	    delInt(1, tuple, -1, Num * 2);
#endif
#ifdef FIELD9
	    f9();
	    delString(1, tuple, -2, &Buffer[0]);
#endif
	    d_print("-");
	    rStep();
	}
#endif
}

xCre(fld)
int fld;
{
    xCreate(1, fld, UNIQUE);
}

#ifdef LOGGING
logger(code, relNum, tuple, fldNum, typ, len, ptr)
    int code;
    word relNum;
    long tuple;
    int fldNum;
    byte typ;
    byte *ptr;
    word len;
{
    d_fwrite(&code, sizeof(int));
    d_fwrite(&relNum, sizeof(word));
    d_fwrite(&tuple, sizeof(long));
    d_fwrite(&fldNum, sizeof(int));
    d_fwrite(&typ, sizeof(byte));
    d_fwrite(&len, sizeof(word));
    d_fwrite(ptr, len);
}
#endif

main()
{
    word i;
#ifdef STATS
    long bb, rd, wrt, get, put;
#endif

    dbInit();
#ifdef DICCRE
    dbCreate("TEST");
    dcField("custno",	FTINT,		0);
    dcField("name",	FTUCSTR,	30);
    dcField("address",	FTSTRING,	40);
    dcField("phone",	FTCHAR,		7);
    dcField("balance",	FTFLOAT,	0);
    dcField("status",	FTBOOL,		0);
    dcField("SSN",	FTLONG,		0);
    dcField("long",	FTSTRING,	MAXKEY);
    dcEnd();
#endif
    dbOpen("TEST");
#ifdef CREATE
	rCreate(1, "CUST", "0", "0", FILESIZE);
	for (i = 1; i <= 7; i++) {
	    rcField(i);
	}
	rcEnd();
	for (i = 1; i <= 7; i++) {
	    xCre(i);
	}
#ifdef FIELD8
	xCre(-1);
#endif
#ifdef FIELD9
	xCre(-2);
#endif
#else
	rOpen(1, '0', "CUST");
#endif

#ifdef LOGGING
    d_fcreate('A', "CUST", "LOG", false);
    setLog(logger);
#endif
    run();
#ifdef RERUN
	if (!rClose(1)) {
	    d_print("rClose failed.\r\n");
	} else {
	    d_print("\r\nStarting second run.\r\n");
	    rOpen(1, '0', "CUST");
	    run();
	}
#endif
#ifdef LOGGING
    d_fwrite("\0\0", 2);
    d_fclose();
#endif
#ifdef STATS
    i = d_blkCnt();
    d_print("\r\nNumber of buffers = ");
    d_pNum(i);
    d_print("\r\n");
    d_bStats(&bb, &rd, &wrt, &get, &put);
    pLong("Number of d_bbGet calls = ", bb);
    pLong("Number of physical reads = ", rd);
    pLong("Number of physical writes = ", wrt);
    pLong("Number of field gets = ", get);
    pLong("Number of field puts = ", put);
#endif
    dbTerm();
}
