#include "system.h"
#include "dbint.h"
#include "fcb.h"

/* file of utility routines. */

int (*d_EHand)();
int (*d_AHand)();

char *d_DBMESS = "*** database error #";

/*
 * d_eInit - initialize for error handling.
 */

d_eInit()
{
    d_EHand = nil;
    d_AHand = nil;
}

/*
 * dbSetError - set up an error handler.
 */

dbSetError(handler)
    int (*handler)();
{
    d_EHand = handler;
}

/*
 * dbSetAbort - set up an abort handler.
 */

dbSetAbort(handler)
    int (*handler)();
{
    d_AHand = handler;
}

/*
 * d_fcbName - fill a CP/M or MSDOS name into the passed FCB.
 *	       Assumption: file "typ" is 3 characters long.
 */

d_fcbName(f, drive, name, typ)
    struct FCB *f;
    char drive;
    char *name;
    char *typ;
{
    register int i;

    if (drive == 0 || drive == '0') {
	f->fcbdrv = 0;
    } else {
	f->fcbdrv = drive - ('A' - 1);
    }
    for (i = 0; i != 8 && *name != '\0';) {
	f->fcbnam[i++] = *name++;
    }
    while (i != 8) {
	f->fcbnam[i++] = ' ';
    }
    for (i = 0; i != 3;) {
	f->fcbext[i++] = *typ++;
    }
}

#ifdef XENIX
/*
 * d_cpmName - assemble a CP/M or MSDOS file name from the given drive,
 *	       base name and extension into the passed buffer.
 */

d_cpmName(buff, drive, name, typ)
    char drive;
    char *buff, *name, *typ;
{
    byte count;

    if (drive != 0 && drive != '0') {
	*buff++ = drive + ('A' - 1);
#ifdef XENIX
	*buff++ = '/';
#else
	*buff++ = ':';
#endif
    }
    count = 8;
    while (count-- != 0 && *name != 0 && *name != ' ') {
	*buff++ = *name++;
    }
    if (*typ != 0) {
	*buff++ = '.';
	count = 3;
	while (count-- != 0 && *typ != 0 && *typ != ' ') {
	    *buff++ = *typ++;
	}
    }
    *buff = 0;
}
#endif

/*
 * d_print - print a string.
 */

d_print(mess)
    register char *mess;
{
#ifdef XENIX
    char *p;
    
    p = mess;
    while (*mess++ != 0) ;
    write(2, p, mess - p - 1);
#else
    while (*mess != 0) {
#ifdef MSDOS
	????
#endif
#ifdef CPM
	bdos(0x02, *mess++);
#endif
    }
#endif
}

/*
 * d_crlf - print the appropriate line terminator.
 */

d_crlf()
{
#ifdef XENIX
    d_print("\n");
#endif
#ifdef MSDOS
    d_print("\r\n");
#endif
#ifdef CPM
    d_print("\r\n");
#endif
}

/*
 * d_pNum - print a number.
 */

d_pNum(n)
    word n;
{
    static char buffer[6];
    register char *p;

    p = &buffer[5];
    *p = '\0';
    do {
	*--p = n % 10 + '0';
	n /= 10;
    } while (n != 0);
    d_print(p);
}

/*
 * d_errEnd - abort with a tail of a message.
 */

d_errEnd()
{
    d_print(". ***");
    d_crlf();
#ifdef XENIX
    abort();
#else
    exit(1);
#endif
}

/*
 * d_setErr - handle a non-fatal error message from a database routine.
 */

d_setErr(code)
    word code;
{
    if (d_EHand != nil) {
	(*d_EHand)(code);
    } else {
	d_print("*** application error #");
	d_pNum(code);
	d_errEnd();
    }
}

/*
 * d_error - handle a fatal error message from a database routine.
 */

d_error(code)
    word code;
{
    d_setErr(code + 1000);
#ifdef XENIX
    abort();
#else
    exit(1);
#endif
}

/*
 * d_abort - print an error code and abort.
 */

d_abort(code)
    word code;
{
    if (d_AHand != nil) {
	(*d_AHand)(code);
#ifdef XENIX
	abort();
#else
	exit(1);
#endif
    } else {
	d_print(d_DBMESS);
	d_pNum(code);
	d_errEnd();
    }
}

/*
 * d_2abort - print an error number and two numbers, then exit.
 */

d_2abort(code, n1, n2)
    word code, n1, n2;
{
    if (d_AHand != nil) {
	(*d_AHand)(code);
#ifdef XENIX
	abort();
#else
	exit(1);
#endif
    } else {
	d_print(d_DBMESS);
	d_pNum(code);
	d_print(": ");
	d_pNum(n1);
	d_print(", ");
	d_pNum(n2);
	d_errEnd();
    }
}

/*
 * d_alloc - alloc space, but check returned pointer.
 */

byte *
d_alloc(size)
    word size;
{
#ifdef XENIX
    byte *malloc();
    byte *ptr;

    ptr = malloc(size);
    if (ptr == nil) {
	d_abort(1);
    }
#endif
#ifdef MSDOS
    byte *getmem();
    byte *ptr;

    ptr = getmem(size);
    if (ptr == nil || ptr == -1) {
	d_abort(1);
    }
#endif
#ifdef CPM
    byte *malloc();
    byte *ptr;

    ptr = alloc(size);
    if (ptr == nil || ptr == -1) {
	d_abort(1);
    }
#endif
    return(ptr);
}

/*
 * d_free - dummy routine to free memory.
 */

d_free(ptr, len)
    byte *ptr;
    word len;
{
#ifdef XENIX
    free(ptr);
#endif
#ifdef MSDOS
    rlsmem(ptr, len);
#endif
#ifdef CPM
    /* Aztec C doesn't have a free. */
#endif
}

/*
 * d_cpPtr - copy a POINTER.
 */

d_cpPtr(dest, source)
    POINTER *dest, *source;
{

    COPY(dest, source, sizeof(POINTER));
}

/*
 * d_ptrEq - return true if two POINTERs are equal.
 */

bool
d_ptrEq(left, right)
    POINTER *left, *right;
{

    return(left->p_file == right->p_file && left->p_block == right->p_block);
}
